"""
OiaFed 命令行接口

提供专业的命令行工具，支持多种操作：
- init: 初始化新实验（生成配置模板）
- generate: 从基础配置生成多客户端配置
- run: 运行联邦学习实验
- validate: 验证配置文件
- list: 列出可用组件
- version: 显示版本信息

使用方式:
    # 初始化实验
    oiafed init my_experiment --algorithm fedavg --dataset cifar10
    oiafed init fcl_exp --scenario fcl --num-clients 5
    
    # 生成配置
    oiafed generate --base base.yaml --num-clients 10
    
    # 运行实验
    oiafed run --config config.yaml
    oiafed run --config configs/ --mode parallel
    
    # 其他
    oiafed validate --config config.yaml
    oiafed list aggregators
    oiafed version
"""
import os
os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"  # 避免 grpcio 与 protobuf 版本冲突问题
import argparse
import sys
import os
import re
from pathlib import Path
from typing import List, Optional


def _load_yaml_with_env(file_path: Path) -> dict:
    """
    加载 YAML 文件并替换环境变量
    
    支持语法: ${VAR_NAME:default_value} 或 ${VAR_NAME}
    
    Args:
        file_path: YAML 文件路径
        
    Returns:
        解析后的字典
    """
    import yaml
    
    with open(file_path, "r", encoding="utf-8") as f:
        content = f.read()
    
    # 替换环境变量 ${VAR:default}
    def replace_env_var(match):
        var_name = match.group(1)
        default_value = match.group(2) if match.group(2) else ""
        return os.environ.get(var_name, default_value)
    
    # 匹配 ${VAR:default} 或 ${VAR}
    content = re.sub(r'\$\{([^}:]+)(?::([^}]*))?\}', replace_env_var, content)
    
    return yaml.safe_load(content) or {}


def get_version() -> str:
    """获取版本号"""
    try:
        from . import __version__
        return __version__
    except ImportError:
        return "0.1.0"


def get_parser() -> argparse.ArgumentParser:
    """创建命令行解析器"""
    parser = argparse.ArgumentParser(
        prog="oiafed",
        description="OiaFed: One Framework for All Federation",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  # 初始化新实验
  oiafed init my_exp --algorithm fedavg --dataset cifar10 --num-clients 10
  oiafed init fcl_exp --scenario fcl --dataset cifar10
  
  # 从基础配置生成多客户端配置
  oiafed generate --base base.yaml --num-clients 10
  oiafed generate --base base.yaml -n 5 --partition dirichlet --alpha 0.3
  
  # 运行实验
  oiafed run --config configs/fedavg.yaml
  oiafed run --config configs/experiment/ --mode parallel
  
  # 验证配置
  oiafed validate --config config.yaml
  
  # 列出组件
  oiafed list aggregators
  oiafed list learners
  
  # 显示版本
  oiafed version

Documentation: https://docs.oiafed.cn
        """
    )
    
    subparsers = parser.add_subparsers(dest="command", help="可用命令")
    
    # ========== run 命令 ==========
    run_parser = subparsers.add_parser(
        "run",
        help="运行联邦学习实验",
        description="从配置文件或论文定义运行联邦学习实验"
    )
    
    # 配置来源
    run_parser.add_argument(
        "--config", "-c",
        type=str,
        default=None,
        help="配置文件/目录路径。如果是目录则直接运行；如果是文件则作为补充配置"
    )
    run_parser.add_argument(
        "--paper", "-p",
        type=str,
        default=None,
        help="论文ID (如 fedavg, fot)，使用论文默认配置"
    )
    
    # 客户端数量（--paper 模式必需）
    run_parser.add_argument(
        "--num-clients", "-n",
        type=int,
        default=None,
        help="客户端数量 (--paper 模式必需)"
    )
    
    # 运行模式
    run_parser.add_argument(
        "--mode", "-m",
        type=str,
        choices=["auto", "serial", "parallel"],
        default="parallel",
        help="运行模式 (default: parallel)"
    )
    
    # 便捷参数 - 覆盖配置
    run_parser.add_argument(
        "--rounds",
        type=int,
        default=None,
        help="训练轮数"
    )
    run_parser.add_argument(
        "--local-epochs",
        type=int,
        default=None,
        help="本地训练轮数"
    )
    run_parser.add_argument(
        "--lr", "--learning-rate",
        type=float,
        default=None,
        dest="learning_rate",
        help="学习率"
    )
    run_parser.add_argument(
        "--batch-size",
        type=int,
        default=None,
        help="批大小"
    )
    run_parser.add_argument(
        "--data-dir",
        type=str,
        default=None,
        help="数据目录"
    )
    run_parser.add_argument(
        "--output-dir", "-o",
        type=str,
        default=None,
        help="输出目录"
    )
    run_parser.add_argument(
        "--seed",
        type=int,
        default=None,
        help="随机种子"
    )
    
    # 其他参数
    run_parser.add_argument(
        "--log-level",
        type=str,
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        default="INFO",
        help="日志级别 (default: INFO)"
    )
    run_parser.add_argument(
        "--dry-run",
        action="store_true",
        help="只显示配置，不实际运行"
    )
    run_parser.add_argument(
        "--save-config",
        type=str,
        default=None,
        help="保存生成的配置到指定目录"
    )
    
    # ========== validate 命令 ==========
    validate_parser = subparsers.add_parser(
        "validate",
        help="验证配置文件",
        description="验证配置文件的正确性"
    )
    validate_parser.add_argument(
        "--config", "-c",
        type=str,
        required=True,
        help="配置文件路径"
    )
    validate_parser.add_argument(
        "--verbose", "-v",
        action="store_true",
        help="显示详细信息"
    )
    
    # ========== list 命令 ==========
    list_parser = subparsers.add_parser(
        "list",
        help="列出可用组件",
        description="列出框架中注册的组件"
    )
    list_parser.add_argument(
        "component_type",
        type=str,
        nargs="?",
        choices=["all", "aggregators", "learners", "models", "datasets", "trainers"],
        default="all",
        help="组件类型 (default: all)"
    )
    
    # ========== show-config 命令 ==========
    show_parser = subparsers.add_parser(
        "show-config",
        help="显示解析后的配置",
        description="显示完整的解析后配置（包括继承和默认值）"
    )
    show_parser.add_argument(
        "--config", "-c",
        type=str,
        required=True,
        help="配置文件路径"
    )
    
    # ========== version 命令 ==========
    subparsers.add_parser(
        "version",
        help="显示版本信息"
    )
    
    # ========== info 命令 ==========
    subparsers.add_parser(
        "info",
        help="显示框架信息"
    )
    
    # ========== init 命令 ==========
    init_parser = subparsers.add_parser(
        "init",
        help="初始化新实验",
        description="生成实验配置模板"
    )
    init_parser.add_argument(
        "name",
        type=str,
        help="实验名称"
    )
    init_parser.add_argument(
        "--algorithm", "-a",
        type=str,
        default="fedavg",
        help="聚合算法 (default: fedavg)"
    )
    init_parser.add_argument(
        "--dataset", "-d",
        type=str,
        default="cifar10",
        choices=["mnist", "fmnist", "cifar10", "cifar100"],
        help="数据集 (default: cifar10)"
    )
    init_parser.add_argument(
        "--num-clients", "-n",
        type=int,
        default=10,
        help="客户端数量 (default: 10)"
    )
    init_parser.add_argument(
        "--output", "-o",
        type=str,
        default="./configs",
        help="输出目录 (default: ./configs)"
    )
    init_parser.add_argument(
        "--scenario",
        type=str,
        default="hfl",
        choices=["hfl", "pfl", "fcl", "vfl", "fu"],
        help="联邦场景 (default: hfl)"
    )
    init_parser.add_argument(
        "--paper", "-p",
        type=str,
        default=None,
        help="从论文生成配置 (如 --paper fedavg)，会覆盖 --algorithm 和 --scenario"
    )
    
    # ========== generate 命令 ==========
    gen_parser = subparsers.add_parser(
        "generate",
        help="生成多客户端配置",
        description="从基础配置生成多个 learner 配置文件"
    )
    gen_parser.add_argument(
        "--base", "-b",
        type=str,
        required=True,
        help="基础配置文件路径"
    )
    gen_parser.add_argument(
        "--num-clients", "-n",
        type=int,
        required=True,
        help="客户端数量"
    )
    gen_parser.add_argument(
        "--output", "-o",
        type=str,
        default=None,
        help="输出目录 (default: 与base同目录)"
    )
    gen_parser.add_argument(
        "--partition",
        type=str,
        default="dirichlet",
        choices=["iid", "dirichlet", "label_skew", "quantity_skew"],
        help="数据划分策略 (default: dirichlet)"
    )
    gen_parser.add_argument(
        "--alpha",
        type=float,
        default=0.5,
        help="Dirichlet alpha参数 (default: 0.5)"
    )
    
    # ========== papers 命令 ==========
    papers_parser = subparsers.add_parser(
        "papers",
        help="论文库管理",
        description="浏览、查询论文实现。运行论文请使用: oiafed run --paper <paper_id> -n <num_clients>"
    )
    papers_subparsers = papers_parser.add_subparsers(dest="papers_command", help="论文操作")
    
    # papers list
    papers_list_parser = papers_subparsers.add_parser(
        "list",
        help="列出所有论文"
    )
    papers_list_parser.add_argument(
        "--category", "-c",
        type=str,
        choices=["HFL", "VFL", "FCL", "FU"],
        help="按类别筛选"
    )
    
    # papers show
    papers_show_parser = papers_subparsers.add_parser(
        "show",
        help="显示论文详情"
    )
    papers_show_parser.add_argument(
        "paper_id",
        type=str,
        help="论文ID"
    )
    papers_show_parser.add_argument(
        "--params", "-p",
        action="store_true",
        help="显示可调参数"
    )
    
    # papers init
    papers_init_parser = papers_subparsers.add_parser(
        "init",
        help="生成论文配置模板"
    )
    papers_init_parser.add_argument(
        "paper_id",
        type=str,
        help="论文ID"
    )
    papers_init_parser.add_argument(
        "--output", "-o",
        type=str,
        default="./configs",
        help="输出目录 (default: ./configs)"
    )
    papers_init_parser.add_argument(
        "--num-clients", "-n",
        type=int,
        help="客户端数量 (覆盖默认值)"
    )
    
    return parser


def cmd_run(args: argparse.Namespace) -> int:
    """执行 run 命令"""
    from .infra.logging import setup_logging
    
    # 设置日志
    setup_logging(node_id="cli", level=args.log_level, console=True)
    
    # 判断运行模式
    config_path = Path(args.config) if args.config else None
    
    # 情况1: --config 是目录，直接运行（原有逻辑）
    if config_path and config_path.is_dir():
        return _run_from_config_dir(args, config_path)
    
    # 情况2: --paper 模式，从论文生成配置运行
    if args.paper:
        return _run_from_paper(args, config_path)
    
    # 情况3: 只有 --config 文件，没有 --paper
    if config_path and config_path.is_file():
        print("错误: 单个配置文件需要配合 --paper 使用，或者指定配置目录", file=sys.stderr)
        print("示例:")
        print("  oiafed run --config ./configs/my_exp/     # 运行配置目录")
        print("  oiafed run --paper fedavg -n 10           # 使用论文默认")
        print("  oiafed run --paper fedavg -n 10 --config base.yaml  # 论文+配置文件")
        return 1
    
    # 情况4: 什么都没指定
    print("错误: 必须指定 --config 或 --paper", file=sys.stderr)
    print("示例:")
    print("  oiafed run --config ./configs/my_exp/     # 运行配置目录")
    print("  oiafed run --paper fedavg -n 10           # 使用论文默认")
    return 1


def _run_from_config_dir(args: argparse.Namespace, config_path: Path) -> int:
    """从配置目录运行（原有逻辑）"""
    from .runner import FederationRunner
    
    if not config_path.exists():
        print(f"错误: 配置目录不存在: {config_path}", file=sys.stderr)
        return 1
    
    try:
        runner = FederationRunner(str(config_path))
        
        if args.dry_run:
            print("配置解析成功！")
            print(f"  节点数: {len(runner.configs)}")
            print(f"  节点ID: {runner.get_node_ids()}")
            print(f"  运行模式: {runner._get_execution_mode()}")
            return 0
        
        # 运行实验
        result = runner.run_sync()
        
        print("\n实验完成！")
        if result:
            print(f"  结果: {result}")
        
        return 0
        
    except Exception as e:
        print(f"错误: {e}", file=sys.stderr)
        if args.log_level == "DEBUG":
            import traceback
            traceback.print_exc()
        return 1


def _run_from_paper(args: argparse.Namespace, config_file: Optional[Path]) -> int:
    """从论文定义运行"""
    import yaml
    import tempfile
    import shutil
    import socket
    
    try:
        from .papers import get_registry
    except ImportError:
        from oiafed.papers import get_registry
    
    # 验证参数
    if args.num_clients is None:
        print("错误: --paper 模式必须指定 -n/--num-clients", file=sys.stderr)
        return 1
    
    registry = get_registry()
    paper = registry.get(args.paper)
    
    if not paper:
        print(f"错误: 未找到论文 '{args.paper}'", file=sys.stderr)
        print("使用 'oiafed papers list' 查看所有可用论文")
        return 1
    
    # 1. 加载论文默认配置
    paper_defaults = registry.get_defaults(args.paper)
    
    # 2. 加载配置文件（如果指定）- 支持环境变量替换
    file_config = {}
    if config_file and config_file.exists():
        file_config = _load_yaml_with_env(config_file)
    elif config_file is None:
        # 尝试加载默认 base.yaml
        default_base = Path("configs/base.yaml")
        if default_base.exists():
            file_config = _load_yaml_with_env(default_base)
            print(f"使用默认配置: {default_base}")
    
    # 3. 构建命令行覆盖参数
    cli_override = {}
    
    if args.rounds is not None:
        cli_override.setdefault("trainer", {})["num_rounds"] = args.rounds
    if args.local_epochs is not None:
        cli_override.setdefault("trainer", {})["local_epochs"] = args.local_epochs
    if args.learning_rate is not None:
        cli_override.setdefault("learner", {})["learning_rate"] = args.learning_rate
    if args.batch_size is not None:
        cli_override.setdefault("learner", {})["batch_size"] = args.batch_size
    if args.data_dir is not None:
        cli_override.setdefault("dataset", {})["data_dir"] = args.data_dir
    if args.seed is not None:
        cli_override["seed"] = args.seed
    
    # 4. 合并配置：paper_defaults < file_config < cli_override
    merged = _deep_merge(paper_defaults, file_config)
    merged = _deep_merge(merged, cli_override)
    
    # 5. 获取运行模式
    mode = args.mode
    if mode == "auto":
        mode = file_config.get("mode", "parallel")
    
    # 6. 生成临时配置目录或保存到指定目录
    if args.save_config:
        temp_dir = args.save_config
        Path(temp_dir).mkdir(parents=True, exist_ok=True)
        cleanup = False
    else:
        temp_dir = tempfile.mkdtemp(prefix=f"oiafed_{args.paper}_")
        cleanup = True
    
    try:
        # 生成配置文件
        _generate_paper_configs(
            paper=paper,
            registry=registry,
            merged_config=merged,
            num_clients=args.num_clients,
            output_dir=temp_dir,
            mode=mode,
            file_config=file_config,
        )
        
        if args.dry_run:
            # 计算 exp_name 和划分信息用于显示
            dataset_type = paper.get_component("dataset") or "cifar10"
            dataset_config = merged.get("dataset", {})
            partition_config = dataset_config.get("partition", {})
            partition_strategy = partition_config.get("strategy", "dirichlet")
            
            if "exp_name" in file_config:
                exp_name = file_config["exp_name"]
            else:
                exp_name = _generate_exp_name(
                    algorithm=paper.id,
                    dataset=dataset_type,
                    partition_strategy=partition_strategy,
                    partition_config=partition_config,
                )
            
            print(f"\n[Dry Run] 论文: {paper.name}")
            print(f"{'=' * 50}")
            print(f"  实验名称: {exp_name}")
            print(f"  客户端数: {args.num_clients}")
            print(f"  运行模式: {mode}")
            print(f"  配置目录: {temp_dir}")
            print(f"")
            print(f"  数据集配置:")
            print(f"    - 数据集: {dataset_type}")
            print(f"    - 划分策略: {partition_strategy}")
            if partition_strategy == "dirichlet":
                print(f"    - alpha: {partition_config.get('alpha', 0.5)}")
            elif partition_strategy == "label_skew":
                print(f"    - num_labels_per_client: {partition_config.get('num_labels_per_client', 2)}")
            print(f"")
            print(f"  组件:")
            for comp_type, comp_name in paper.components.items():
                print(f"    - {comp_type}: {comp_name}")
            print(f"")
            print(f"  训练参数:")
            print(f"    - rounds: {merged.get('trainer', {}).get('num_rounds', 'N/A')}")
            print(f"    - local_epochs: {merged.get('trainer', {}).get('local_epochs', 'N/A')}")
            print(f"    - learning_rate: {merged.get('learner', {}).get('learning_rate', 'N/A')}")
            print(f"    - batch_size: {merged.get('learner', {}).get('batch_size', 'N/A')}")
            print(f"")
            # 显示 tracker 配置
            tracker_config = merged.get("tracker", {})
            if tracker_config.get("enabled", False) or tracker_config.get("backends"):
                print(f"  实验追踪:")
                for backend in tracker_config.get("backends", []):
                    backend_type = backend.get("type", "unknown")
                    print(f"    - {backend_type}")
                    if backend_type == "mlflow":
                        args_dict = backend.get("args", {})
                        if args_dict.get("tracking_uri"):
                            print(f"      tracking_uri: {args_dict['tracking_uri']}")
                        if args_dict.get("experiment_name"):
                            print(f"      experiment_name: {args_dict['experiment_name']}")
                print(f"")
            print(f"{'=' * 50}")
            
            if args.save_config:
                print(f"\n配置已保存到: {temp_dir}")
            return 0
        
        # 运行实验
        # 计算 exp_name 用于显示
        dataset_type = paper.get_component("dataset") or "cifar10"
        dataset_config = merged.get("dataset", {})
        partition_config = dataset_config.get("partition", {})
        partition_strategy = partition_config.get("strategy", "dirichlet")
        
        if "exp_name" in file_config:
            exp_name = file_config["exp_name"]
        else:
            exp_name = _generate_exp_name(
                algorithm=paper.id,
                dataset=dataset_type,
                partition_strategy=partition_strategy,
                partition_config=partition_config,
            )
        
        print(f"\n运行论文实验: {paper.name}")
        print(f"  实验名称: {exp_name}")
        print(f"  客户端数: {args.num_clients}")
        print(f"  运行模式: {mode}")
        print(f"{'=' * 50}")
        
        from .runner import FederationRunner
        runner = FederationRunner(temp_dir)
        result = runner.run_sync()
        
        print(f"\n实验完成！")
        return 0
        
    except Exception as e:
        print(f"错误: {e}", file=sys.stderr)
        if args.log_level == "DEBUG":
            import traceback
            traceback.print_exc()
        return 1
    finally:
        # 清理临时目录
        if cleanup and not args.dry_run:
            shutil.rmtree(temp_dir, ignore_errors=True)


def _generate_paper_configs(
    paper,
    registry,
    merged_config: dict,
    num_clients: int,
    output_dir: str,
    mode: str,
    file_config: dict,
) -> None:
    """生成论文配置文件"""
    import yaml
    import socket
    
    output_path = Path(output_dir)
    
    # 获取网络配置
    network_config = file_config.get("network", {})
    trainer_port = network_config.get("trainer_port", 50051)
    learner_base_port = network_config.get("learner_base_port", 50052)
    auto_find_port = network_config.get("auto_find_port", True)
    
    # 获取数据集和划分配置
    dataset_type = paper.get_component("dataset") or "cifar10"
    dataset_config = merged_config.get("dataset", {})
    partition_config = dataset_config.get("partition", {})
    partition_strategy = partition_config.get("strategy", "dirichlet")
    
    # 生成 exp_name: {algorithm}_{dataset}_{partition}
    # 如果用户显式指定了 exp_name，则使用用户指定的
    if "exp_name" in file_config:
        exp_name = file_config["exp_name"]
    else:
        exp_name = _generate_exp_name(
            algorithm=paper.id,
            dataset=dataset_type,
            partition_strategy=partition_strategy,
            partition_config=partition_config,
        )
    
    data_dir = dataset_config.get("data_dir", "./data")
    seed = merged_config.get("seed", 42)
    
    # 生成统一的 run_name（时间戳），确保所有节点使用相同的值
    from datetime import datetime
    run_name = file_config.get("run_name") or datetime.now().strftime("%Y%m%d_%H%M%S")
    
    # 日志和追踪配置（深度合并）
    # 确保 exp_name 和 run_name 传递到 logging 配置
    logging_config = merged_config.get("logging", {})
    logging_config = {
        "level": logging_config.get("level", "INFO"),
        "console": logging_config.get("console", True),
        "exp_name": exp_name,      # 使用生成的实验名称
        "run_name": run_name,      # 使用统一的运行名称
        **{k: v for k, v in logging_config.items() if k not in ["level", "console", "exp_name", "run_name"]},
    }
    tracker_config = merged_config.get("tracker", {})
    
    # 如果需要，查找可用端口
    if mode == "parallel" and auto_find_port:
        trainer_port = _find_available_port(trainer_port)
        # 确保 learner_base_port 从 trainer_port 之后开始，避免端口冲突
        learner_base_port = _find_available_port(max(learner_base_port, trainer_port + 1))
    
    # ===== 生成 Trainer 配置 =====
    trainer_config = {
        "node_id": "trainer",
        "role": "trainer",
        "exp_name": exp_name,
        
        "listen": {
            "host": "localhost",
            "port": trainer_port,
        },
        "min_peers": num_clients,
        
        "trainer": {
            "type": paper.get_component("trainer") or "default",
            "args": merged_config.get("trainer", {}),
        },
        
        "aggregator": {
            "type": paper.get_component("aggregator") or "fedavg",
            "args": merged_config.get("aggregator", {}),
        },
        
        "model": {
            "type": paper.get_component("model") or "cnn",
            "args": merged_config.get("model", {}),
        },
        
        "logging": logging_config,
        "transport": {"mode": "grpc" if mode == "parallel" else "memory"},
    }
    
    if tracker_config.get("enabled", True) and tracker_config.get("backends"):
        trainer_config["tracker"] = tracker_config
    
    trainer_path = output_path / "trainer.yaml"
    with open(trainer_path, "w", encoding="utf-8") as f:
        yaml.dump(trainer_config, f, default_flow_style=False, allow_unicode=True)
    
    # ===== 生成 Learner 配置 =====
    # 注意：dataset_type 和 partition_config 已在上面获取
    # 这里需要复制 dataset_config 以避免修改原始配置
    dataset_config_copy = {k: v for k, v in dataset_config.items() if k != "partition"}
    
    for i in range(num_clients):
        learner_port = learner_base_port + i
        if mode == "parallel" and auto_find_port and i > 0:
            learner_port = _find_available_port(learner_port)
        
        learner_config = {
            "node_id": f"learner_{i}",
            "role": "learner",
            "exp_name": exp_name,
            
            "listen": {
                "host": "localhost",
                "port": learner_port,
            },
            "connect_to": [f"trainer@localhost:{trainer_port}"],
            
            "learner": {
                "type": paper.get_component("learner") or "default",
                "args": merged_config.get("learner", {}),
            },
            
            "model": {
                "type": paper.get_component("model") or "cnn",
                "args": merged_config.get("model", {}),
            },
            
            "datasets": [
                {
                    "type": dataset_type,
                    "split": "train",
                    "args": {
                        "data_dir": data_dir,
                        "download": True,
                        **{k: v for k, v in dataset_config_copy.items() if k not in ["data_dir", "download"]},
                    },
                    "partition": {
                        "strategy": partition_strategy,
                        "num_partitions": num_clients,
                        "partition_id": i,
                        "seed": seed,
                        **{k: v for k, v in partition_config.items() if k not in ["strategy", "num_partitions", "partition_id", "seed"]},
                    },
                },
                {
                    "type": dataset_type,
                    "split": "test",
                    "args": {
                        "data_dir": data_dir,
                    },
                },
            ],
            
            "logging": logging_config,
            "transport": {"mode": "grpc" if mode == "parallel" else "memory"},
        }
        
        # Learner 也需要 tracker 配置（用于同步后记录指标）
        if tracker_config.get("enabled", True) and tracker_config.get("backends"):
            learner_config["tracker"] = tracker_config
        
        learner_path = output_path / f"learner_{i}.yaml"
        with open(learner_path, "w", encoding="utf-8") as f:
            yaml.dump(learner_config, f, default_flow_style=False, allow_unicode=True)


def _find_available_port(start_port: int, max_attempts: int = 100) -> int:
    """查找可用端口"""
    import socket
    
    for i in range(max_attempts):
        port = start_port + i
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind(("localhost", port))
                return port
        except OSError:
            continue
    
    # 如果找不到，返回原端口（让后续程序报错）
    return start_port


def _generate_exp_name(
    algorithm: str,
    dataset: str,
    partition_strategy: str,
    partition_config: dict,
) -> str:
    """
    自动生成实验名称
    
    格式: {algorithm}_{dataset}_{partition}
    
    示例:
    - fedavg_cifar10_iid
    - fedavg_cifar10_dir_0.5
    - moon_mnist_label_skew_3
    - scaffold_cifar100_quantity_skew_0.8
    
    Args:
        algorithm: 算法名称 (paper.id)
        dataset: 数据集名称
        partition_strategy: 划分策略 (iid, dirichlet, label_skew, quantity_skew)
        partition_config: 划分配置字典
        
    Returns:
        生成的实验名称
    """
    # 基础名称
    parts = [algorithm, dataset]
    
    # 根据划分策略添加后缀
    if partition_strategy == "iid":
        parts.append("iid")
    elif partition_strategy == "dirichlet":
        alpha = partition_config.get("alpha", 0.5)
        parts.append(f"dir_{alpha}")
    elif partition_strategy == "label_skew":
        num_labels = partition_config.get("num_labels_per_client", 2)
        parts.append(f"label_skew_{num_labels}")
    elif partition_strategy == "quantity_skew":
        imbalance = partition_config.get("imbalance_ratio", 0.5)
        parts.append(f"quantity_skew_{imbalance}")
    else:
        # 未知策略，直接使用策略名
        parts.append(partition_strategy)
    
    return "_".join(parts)


def _deep_merge(base: dict, override: dict) -> dict:
    """深度合并字典"""
    import copy
    result = copy.deepcopy(base)
    
    for key, value in override.items():
        if key in result and isinstance(result[key], dict) and isinstance(value, dict):
            result[key] = _deep_merge(result[key], value)
        else:
            result[key] = copy.deepcopy(value)
    
    return result


def cmd_validate(args: argparse.Namespace) -> int:
    """执行 validate 命令"""
    from .config import ConfigManager
    
    config_path = Path(args.config)
    if not config_path.exists():
        print(f"错误: 配置文件不存在: {config_path}", file=sys.stderr)
        return 1
    
    try:
        manager = ConfigManager()
        configs = manager.load(str(config_path))
        
        print(f"✓ 配置验证通过: {config_path}")
        
        if args.verbose:
            print(f"\n解析结果:")
            for config in configs:
                print(f"  - {config.node_id} ({config.role})")
                print(f"    exp_name: {config.exp_name}")
                if hasattr(config, 'trainer') and config.trainer:
                    print(f"    trainer: {config.trainer.get('name', 'default')}")
                if hasattr(config, 'aggregator') and config.aggregator:
                    print(f"    aggregator: {config.aggregator.get('name', 'fedavg')}")
        
        return 0
        
    except Exception as e:
        print(f"✗ 配置验证失败: {e}", file=sys.stderr)
        return 1


def cmd_list(args: argparse.Namespace) -> int:
    """执行 list 命令"""
    # 导入 methods 模块以触发所有组件注册
    from . import methods
    # 使用全局 registry（已注册了组件），而不是创建新实例
    from .registry import registry
    
    component_map = {
        "aggregators": "aggregator",
        "learners": "learner",
        "models": "model",
        "datasets": "dataset",
        "trainers": "trainer",
    }
    
    if args.component_type == "all":
        types_to_show = list(component_map.keys())
    else:
        types_to_show = [args.component_type]
    
    for comp_type in types_to_show:
        prefix = component_map.get(comp_type, comp_type)
        components = registry.list(prefix)
        
        print(f"\n{comp_type.upper()}:")
        print("-" * 40)
        
        if components:
            for namespace in sorted(components):
                # 从 namespace 中提取名称，如 "model.cifar10_cnn" -> "cifar10_cnn"
                name = namespace.split(".")[-1] if "." in namespace else namespace
                # 尝试获取组件的描述信息
                try:
                    cls = registry.get(namespace)
                    desc = getattr(cls, '__doc__', '') or ''
                    desc = desc.split('\n')[0].strip() if desc else ''
                    if desc:
                        print(f"  {name}: {desc[:60]}")
                    else:
                        print(f"  {name}")
                except:
                    print(f"  {name}")
        else:
            print("  (无)")
    
    return 0


def cmd_show_config(args: argparse.Namespace) -> int:
    """执行 show-config 命令"""
    import yaml
    from .config import ConfigManager
    
    config_path = Path(args.config)
    if not config_path.exists():
        print(f"错误: 配置文件不存在: {config_path}", file=sys.stderr)
        return 1
    
    try:
        manager = ConfigManager()
        configs = manager.load(str(config_path))
        
        for config in configs:
            print(f"\n{'='*50}")
            print(f"Node: {config.node_id}")
            print('='*50)
            # 转换为字典并打印
            config_dict = config.to_dict() if hasattr(config, 'to_dict') else vars(config)
            print(yaml.dump(config_dict, default_flow_style=False, allow_unicode=True))
        
        return 0
        
    except Exception as e:
        print(f"错误: {e}", file=sys.stderr)
        return 1


def cmd_version(args: argparse.Namespace) -> int:
    """执行 version 命令"""
    print(f"OiaFed version {get_version()}")
    return 0


def cmd_info(args: argparse.Namespace) -> int:
    """执行 info 命令"""
    import platform
    
    print("OiaFed - One Framework for All Federation")
    print("=" * 50)
    print(f"Version:    {get_version()}")
    print(f"Python:     {platform.python_version()}")
    print(f"Platform:   {platform.platform()}")
    print()
    print("Supported Scenarios:")
    print("  - HFL: Horizontal Federated Learning")
    print("  - VFL: Vertical Federated Learning")
    print("  - FCL: Federated Continual Learning")
    print("  - PFL: Personalized Federated Learning")
    print("  - FU:  Federated Unlearning")
    print()
    print("Documentation: https://docs.oiafed.cn")
    print("Repository:    https://github.com/oiafed/oiafed")
    
    return 0


def _init_from_paper(args: argparse.Namespace, output_dir: Path) -> int:
    """从论文生成配置"""
    import yaml
    
    try:
        from .papers import get_registry
    except ImportError:
        from oiafed.papers import get_registry
    
    registry = get_registry()
    paper = registry.get(args.paper)
    
    if not paper:
        print(f"错误: 未找到论文 '{args.paper}'", file=sys.stderr)
        print(f"使用 'oiafed papers list' 查看所有可用论文")
        return 1
    
    # 生成配置文件
    num_clients = args.num_clients
    configs = registry.generate_node_configs(
        args.paper,
        num_clients=num_clients,
        output_dir=str(output_dir),
    )
    
    # 生成覆盖文件模板
    defaults = registry.get_defaults(args.paper)
    
    override_content = f"""# {paper.name}
# 用户覆盖配置 - 修改后使用 --override 参数运行
# 
# 运行方式:
#   oiafed papers run {args.paper} --override {output_dir}/override.yaml
#   或
#   oiafed run --config {output_dir}/

# 引用论文（可选，用于配置文件联动）
paper: {args.paper}

# 覆盖参数（取消注释并修改需要的参数）
# learner:
#   learning_rate: {defaults.get('learner', {}).get('learning_rate', 0.01)}
#   batch_size: {defaults.get('learner', {}).get('batch_size', 32)}

# trainer:
#   num_rounds: {defaults.get('trainer', {}).get('num_rounds', 100)}
#   local_epochs: {defaults.get('trainer', {}).get('local_epochs', 5)}

# experiment:
#   num_clients: {num_clients}
"""
    
    override_path = output_dir / "override.yaml"
    override_path.write_text(override_content)
    
    print(f"\n✓ 从论文 '{paper.name}' 生成配置")
    print(f"  目录: {output_dir}/")
    print(f"")
    print(f"  生成的文件:")
    for config in configs:
        print(f"    - {config['node_id']}.yaml")
    print(f"    - override.yaml (覆盖模板)")
    print(f"")
    print(f"  组件配置:")
    for comp_type, comp_name in paper.components.items():
        print(f"    - {comp_type}: {comp_name}")
    print(f"")
    print(f"  运行实验:")
    print(f"    oiafed run --config {output_dir}/")
    print(f"    或")
    print(f"    oiafed papers run {args.paper} --override {output_dir}/override.yaml")
    
    return 0


def cmd_init(args: argparse.Namespace) -> int:
    """执行 init 命令 - 生成实验配置模板"""
    import os
    
    output_dir = Path(args.output) / args.name
    output_dir.mkdir(parents=True, exist_ok=True)
    
    # 如果指定了 --paper，使用论文注册表生成配置
    if args.paper:
        return _init_from_paper(args, output_dir)
    
    # 数据集配置
    dataset_configs = {
        "mnist": {"num_classes": 10, "model": "mnist_cnn"},
        "fmnist": {"num_classes": 10, "model": "mnist_cnn"},
        "cifar10": {"num_classes": 10, "model": "cifar10_cnn"},
        "cifar100": {"num_classes": 100, "model": "resnet18"},
    }
    
    ds_config = dataset_configs.get(args.dataset, dataset_configs["cifar10"])
    
    # 场景特定配置
    scenario_configs = {
        "hfl": {
            "trainer_type": "default",
            "learner_type": "default",
        },
        "pfl": {
            "trainer_type": "default",
            "learner_type": "moon",
            "learner_args": "temperature: 0.5\n        mu: 1.0",
        },
        "fcl": {
            "trainer_type": "continual",
            "learner_type": "cl.fot",
            "learner_args": "num_tasks: 5\n        classes_per_task: 2\n        orthogonal_weight: 1.0",
        },
        "vfl": {
            "trainer_type": "default",
            "learner_type": "vfl.splitnn",
            "learner_args": "split_layer: 2",
        },
        "fu": {
            "trainer_type": "default",
            "learner_type": "default",
            "aggregator_type": "faderaser",
            "aggregator_args": "history_dir: ./fed_history\n    record_history: true",
        },
    }
    
    sc_config = scenario_configs.get(args.scenario, scenario_configs["hfl"])
    
    # 生成 trainer 配置
    trainer_config = f'''# OiaFed 实验配置
# 生成命令: oiafed init {args.name} --algorithm {args.algorithm} --dataset {args.dataset} --scenario {args.scenario}

exp_name: {args.name}
node_id: trainer
role: trainer

# ==================== 训练配置 ====================
trainer:
  type: {sc_config.get("trainer_type", "default")}
  args:
    max_rounds: 100
    local_epochs: 5
    client_fraction: 1.0
    eval_every: 10

# ==================== 聚合配置 ====================
aggregator:
  type: {sc_config.get("aggregator_type", args.algorithm)}
  args:
    {sc_config.get("aggregator_args", "weighted: true")}

# ==================== 学习器配置 ====================
learner:
  type: {sc_config.get("learner_type", "default")}
  args:
    batch_size: 64
    learning_rate: 0.01
    {sc_config.get("learner_args", "")}

# ==================== 模型配置 ====================
model:
  type: {ds_config["model"]}
  args:
    num_classes: {ds_config["num_classes"]}

# ==================== 数据配置 ====================
datasets:
  - type: {args.dataset}
    split: train
    args:
      data_dir: ./data
      download: true
    partition:
      strategy: dirichlet
      num_partitions: {args.num_clients}
      config:
        alpha: 0.5
        seed: 42

  - type: {args.dataset}
    split: test
    args:
      data_dir: ./data

# ==================== 追踪配置 ====================
tracker:
  backends:
    - type: mlflow
      tracking_uri: ./mlruns
      experiment_name: {args.name}

# ==================== 运行命令 ====================
# oiafed run --config {output_dir}/trainer.yaml
'''

    # 写入 trainer 配置
    trainer_path = output_dir / "trainer.yaml"
    trainer_path.write_text(trainer_config)
    
    # 生成 learner 配置
    for i in range(args.num_clients):
        learner_config = f'''# Learner {i} 配置
# 继承自 trainer.yaml

exp_name: {args.name}
node_id: learner_{i}
role: learner

learner:
  type: {sc_config.get("learner_type", "default")}
  args:
    batch_size: 64
    learning_rate: 0.01
    {sc_config.get("learner_args", "")}

model:
  type: {ds_config["model"]}
  args:
    num_classes: {ds_config["num_classes"]}
'''
        learner_path = output_dir / f"learner_{i}.yaml"
        learner_path.write_text(learner_config)
    
    # 生成 base.yaml (用于 generate 命令)
    base_config = f'''# 基础配置 - 用于 oiafed generate 命令
exp_name: {args.name}

trainer:
  type: {sc_config.get("trainer_type", "default")}
  args:
    max_rounds: 100
    local_epochs: 5

aggregator:
  type: {sc_config.get("aggregator_type", args.algorithm)}

learner:
  type: {sc_config.get("learner_type", "default")}
  args:
    batch_size: 64
    learning_rate: 0.01

model:
  type: {ds_config["model"]}
  args:
    num_classes: {ds_config["num_classes"]}

dataset:
  type: {args.dataset}
  partition:
    strategy: dirichlet
    alpha: 0.5
'''
    base_path = output_dir / "base.yaml"
    base_path.write_text(base_config)
    
    print(f"✓ 实验配置已生成: {output_dir}/")
    print(f"  - trainer.yaml")
    for i in range(args.num_clients):
        print(f"  - learner_{i}.yaml")
    print(f"  - base.yaml")
    print()
    print(f"运行实验:")
    print(f"  oiafed run --config {output_dir}/trainer.yaml")
    print()
    print(f"或使用整个目录:")
    print(f"  oiafed run --config {output_dir}/")
    
    return 0


def cmd_generate(args: argparse.Namespace) -> int:
    """执行 generate 命令 - 从基础配置生成多客户端配置"""
    import yaml
    
    base_path = Path(args.base)
    if not base_path.exists():
        print(f"错误: 基础配置文件不存在: {base_path}", file=sys.stderr)
        return 1
    
    # 确定输出目录
    if args.output:
        output_dir = Path(args.output)
    else:
        output_dir = base_path.parent
    
    output_dir.mkdir(parents=True, exist_ok=True)
    
    # 读取基础配置（支持环境变量替换）
    base_config = _load_yaml_with_env(base_path)
    
    exp_name = base_config.get('exp_name', 'experiment')
    
    # 生成 trainer 配置
    trainer_config = {
        **base_config,
        'node_id': 'trainer',
        'role': 'trainer',
    }
    
    # 添加数据划分配置
    if 'datasets' not in trainer_config:
        dataset_type = base_config.get('dataset', {}).get('type', 'cifar10')
        partition_strategy = args.partition
        partition_config = {'seed': 42}
        
        if partition_strategy == 'dirichlet':
            partition_config['alpha'] = args.alpha
        
        trainer_config['datasets'] = [
            {
                'type': dataset_type,
                'split': 'train',
                'args': {'data_dir': './data', 'download': True},
                'partition': {
                    'strategy': partition_strategy,
                    'num_partitions': args.num_clients,
                    'config': partition_config,
                }
            },
            {
                'type': dataset_type,
                'split': 'test',
                'args': {'data_dir': './data'},
            }
        ]
    
    trainer_path = output_dir / "trainer.yaml"
    with open(trainer_path, 'w', encoding='utf-8') as f:
        yaml.dump(trainer_config, f, default_flow_style=False, allow_unicode=True)
    
    # 生成 learner 配置
    learner_files = []
    for i in range(args.num_clients):
        learner_config = {
            'exp_name': exp_name,
            'node_id': f'learner_{i}',
            'role': 'learner',
        }
        
        # 复制 learner 和 model 配置
        if 'learner' in base_config:
            learner_config['learner'] = base_config['learner']
        if 'model' in base_config:
            learner_config['model'] = base_config['model']
        
        learner_path = output_dir / f"learner_{i}.yaml"
        with open(learner_path, 'w', encoding='utf-8') as f:
            yaml.dump(learner_config, f, default_flow_style=False, allow_unicode=True)
        learner_files.append(f"learner_{i}.yaml")
    
    print(f"✓ 配置文件已生成: {output_dir}/")
    print(f"  - trainer.yaml")
    for lf in learner_files:
        print(f"  - {lf}")
    print()
    print(f"数据划分: {args.partition}" + (f" (alpha={args.alpha})" if args.partition == "dirichlet" else ""))
    print(f"客户端数: {args.num_clients}")
    print()
    print(f"运行实验:")
    print(f"  oiafed run --config {output_dir}/")
    
    return 0


def cmd_papers(args: argparse.Namespace) -> int:
    """执行 papers 命令"""
    try:
        from .papers import get_registry
    except ImportError:
        from oiafed.papers import get_registry
    
    registry = get_registry()
    
    # 子命令分发
    if args.papers_command == "list":
        return cmd_papers_list(args, registry)
    elif args.papers_command == "show":
        return cmd_papers_show(args, registry)
    elif args.papers_command == "init":
        return cmd_papers_init(args, registry)
    else:
        # 没有子命令，显示帮助
        print(registry.format_paper_list())
        print()
        print("运行论文实验请使用:")
        print("  oiafed run --paper <paper_id> -n <num_clients>")
        print()
        print("示例:")
        print("  oiafed run --paper fedavg -n 10")
        print("  oiafed run --paper fedavg -n 10 --config base.yaml")
        print("  oiafed run --paper fedavg -n 10 --rounds 50 --lr 0.01")
        return 0


def cmd_papers_list(args: argparse.Namespace, registry) -> int:
    """列出论文"""
    if args.category:
        papers = registry.list_by_category(args.category)
        if not papers:
            print(f"类别 {args.category} 没有论文")
            return 0
        
        category_name = registry.CATEGORY_NAMES.get(args.category, args.category)
        print(f"\n[{args.category}] {category_name}")
        print("-" * 40)
        for paper_id in papers:
            paper = registry.get(paper_id)
            print(f"  {paper_id:<15} {paper.name}")
        print()
    else:
        print(registry.format_paper_list())
    
    return 0


def cmd_papers_show(args: argparse.Namespace, registry) -> int:
    """显示论文详情"""
    paper = registry.get(args.paper_id)
    if not paper:
        print(f"错误: 未找到论文 '{args.paper_id}'", file=sys.stderr)
        print(f"使用 'oiafed papers list' 查看所有可用论文")
        return 1
    
    print(registry.format_paper_info(args.paper_id, show_params=args.params))
    return 0


def cmd_papers_init(args: argparse.Namespace, registry) -> int:
    """生成论文配置模板"""
    import yaml
    
    paper = registry.get(args.paper_id)
    if not paper:
        print(f"错误: 未找到论文 '{args.paper_id}'", file=sys.stderr)
        return 1
    
    output_dir = Path(args.output) / args.paper_id
    output_dir.mkdir(parents=True, exist_ok=True)
    
    # 生成配置文件
    num_clients = args.num_clients
    configs = registry.generate_node_configs(
        args.paper_id,
        num_clients=num_clients,
        output_dir=str(output_dir),
    )
    
    # 生成覆盖文件模板
    override_template = {
        "# 用户覆盖配置": "修改此文件后使用 --override 参数运行",
        "learner": {
            "learning_rate": paper.get_default("learner.learning_rate", 0.01),
            "batch_size": paper.get_default("learner.batch_size", 32),
        },
        "trainer": {
            "num_rounds": paper.get_default("trainer.num_rounds", 20),
        },
    }
    
    override_path = output_dir / "override.yaml"
    with open(override_path, "w", encoding="utf-8") as f:
        f.write(f"# {paper.name}\n")
        f.write(f"# 用户覆盖配置 - 修改后使用 --override 参数运行\n\n")
        yaml.dump(override_template, f, default_flow_style=False, allow_unicode=True)
    
    print(f"\n✓ 配置已生成: {output_dir}/")
    print(f"")
    print(f"  生成的文件:")
    for config in configs:
        print(f"    - {config['node_id']}.yaml")
    print(f"    - override.yaml (覆盖模板)")
    print(f"")
    print(f"  运行实验:")
    print(f"    oiafed run --config {output_dir}/")
    print(f"")
    print(f"  或使用 papers run:")
    print(f"    oiafed papers run {args.paper_id} --override {override_path}")
    
    return 0


def main(argv: Optional[List[str]] = None) -> int:
    """CLI 主入口"""
    parser = get_parser()
    args = parser.parse_args(argv)
    
    if args.command is None:
        parser.print_help()
        return 0
    
    # 命令分发
    commands = {
        "run": cmd_run,
        "validate": cmd_validate,
        "list": cmd_list,
        "show-config": cmd_show_config,
        "version": cmd_version,
        "info": cmd_info,
        "init": cmd_init,
        "generate": cmd_generate,
        "papers": cmd_papers,
    }
    
    handler = commands.get(args.command)
    if handler:
        return handler(args)
    else:
        parser.print_help()
        return 1


if __name__ == "__main__":
    sys.exit(main())


"""
联邦学习系统容器

提供统一的系统初始化和管理

设计原则：
- 延迟连接：initialize() 不阻塞，连接在 run() 入口建立
- 支持多种运行模式：本地串行、本地并行、分布式
- 只使用配置类，不使用字典操作
- 运行时信息同步：Trainer 生成统一的 run_name，Learner 连接后同步
"""

import asyncio
import sys
from datetime import datetime
from typing import Any, Dict, List, Optional

from ..config import (
    NodeConfig,
    LogConfig,
    TrackerConfig,
    ConnectionRetryConfig,
    ComponentConfig,
)
from ..proxy import ProxyCollection
from .node import Node
from ..registry import registry
from ..callback import CallbackManager
from ..tracker import CompositeTracker
from ..infra.logging import setup_logging, get_logger, cleanup_logging


class FederatedSystem:
    """
    联邦学习系统容器

    职责：
    1. 统一节点管理（不区分角色）
    2. 配置驱动组件创建（根据配置中存在的组件来初始化）
    3. 依赖注入
    4. 提供统一入口

    设计原则：
    - 只接受 NodeConfig，不支持字典配置
    - 延迟连接：initialize() 只做本地准备，不等待连接
    - 连接在 run() 入口建立，支持重试

    Example:
        from oiafed import load_config, FederatedSystem

        config = load_config("configs/trainer.yaml")
        system = FederatedSystem(config)
        await system.initialize()
        result = await system.run()
        await system.stop()
    """

    def __init__(self, config: NodeConfig):
        """
        初始化系统

        Args:
            config: NodeConfig 实例
        """
        self._config = config

        # Node 将在 initialize 中创建
        self.node: Optional[Node] = None

        # 组件容器
        self.trainer = None
        self.learner = None
        self.learners: Optional[ProxyCollection] = None

        # 基础设施
        self.tracker: Optional[CompositeTracker] = None
        self.callbacks: Optional[CallbackManager] = None

        # 系统日志
        self.sys_logger = None

        # ComponentBuilder
        from ..builder import ComponentBuilder
        self.builder = ComponentBuilder()

        # 延迟连接状态
        self._connected: bool = False
        self._initialized: bool = False

        # Shutdown 状态
        self._shutdown_event: Optional[asyncio.Event] = None
        
        # ===== 运行时信息（用于分布式同步）=====
        # Trainer 会生成这些信息，Learner 会在连接后同步获取
        self._runtime_run_name: Optional[str] = None
        self._runtime_mlflow_run_id: Optional[str] = None
        self._runtime_synced: bool = False  # Learner 是否已同步

    # ========== 属性访问器 ==========

    @property
    def config(self) -> NodeConfig:
        """获取配置对象"""
        return self._config

    @property
    def node_id(self) -> str:
        """获取节点 ID"""
        return self._config.node_id

    @property
    def exp_name(self) -> str:
        """获取实验名称"""
        return self._config.exp_name

    @property
    def run_name(self) -> Optional[str]:
        """获取运行名称"""
        return self._config.run_name

    @property
    def log_dir(self) -> str:
        """获取日志目录"""
        return self._config.log_dir

    # ========== 工厂方法 ==========

    @classmethod
    def from_node_config(cls, config: NodeConfig) -> "FederatedSystem":
        """
        从 NodeConfig 创建系统

        Args:
            config: NodeConfig 实例

        Returns:
            FederatedSystem 实例
        """
        return cls(config)

    # ========== 核心生命周期方法 ==========

    async def initialize(self):
        """
        初始化系统（不阻塞）

        此方法只做本地准备工作，不等待任何远程连接：
        1. 初始化日志系统
        2. 创建并启动 Node
        3. 初始化基础设施（Tracker/Callbacks）
        4. 初始化 Learner 组件并 bind
        5. [Trainer] 注册运行时信息同步方法
        """
        try:
            if self._initialized:
                if self.sys_logger:
                    self.sys_logger.warning(f"系统已初始化: {self.node_id}")
                return

            # 0. [Trainer] 生成统一的 run_name
            if self._config.is_trainer():
                self._generate_runtime_info()

            # 1. 初始化日志系统
            self._setup_logging()
            self.sys_logger = get_logger(self.node_id, "system")
            self.sys_logger.info(f"开始初始化联邦学习节点: node_id={self.node_id}")
            self.sys_logger.info(f"实验: {self.exp_name}")
            if self.run_name:
                self.sys_logger.info(f"运行: {self.run_name}")
            self.sys_logger.info(f"日志目录: {self.log_dir}")

            # 2. 创建 Node
            self.sys_logger.info(f"正在创建通信节点: {self.node_id}")
            await self._create_node()
            self.sys_logger.debug(f"通信节点创建完成: {self.node_id}")

            # 3. 注册事件处理器
            self.sys_logger.debug("注册连接事件处理器")
            self.node.on("connect", self._on_peer_connect)
            self.node.on("disconnect", self._on_peer_disconnect)

            # 4. 启动 Node
            self.sys_logger.debug("启动通信节点")
            await self.node.start()
            self.sys_logger.debug("通信节点已启动")

            # 5. 初始化基础设施
            self.sys_logger.debug("初始化基础设施")
            await self._initialize_infrastructure()

            # 6. 初始化 Learner
            if self._config.learner:
                self.sys_logger.debug("初始化 Learner 组件")
                await self._initialize_learner()

            # 7. [Trainer] 注册运行时信息同步方法
            if self._config.is_trainer():
                self._register_runtime_sync_handler()

            self._initialized = True
            self.sys_logger.info(f"联邦学习系统初始化完成: node_id={self.node_id}")

        except Exception as e:
            if self.sys_logger:
                self.sys_logger.exception(f"系统初始化失败: {e}")
            else:
                import traceback
                print(f"ERROR: 系统初始化失败 (node_id={self.node_id}): {e}", file=sys.stderr)
                traceback.print_exc()
            raise

    async def run(self) -> Dict[str, Any]:
        """
        运行系统

        Returns:
            运行结果字典
        """
        if not self._initialized:
            raise RuntimeError("System not initialized. Call initialize() first.")

        try:
            # 1. 确保连接已建立
            await self._ensure_connected()

            # 2. [Learner] 同步运行时信息（从 Trainer 获取 run_name 和 mlflow_run_id）
            if self._config.is_learner() and not self._config.is_trainer():
                await self._sync_runtime_info_from_trainer()

            # 3. 初始化 Trainer（需要连接信息）
            if self._config.trainer and self.trainer is None:
                self.sys_logger.info(f"[{self.node_id}] 开始初始化 Trainer")
                await self._initialize_trainer()
                self.sys_logger.info(f"[{self.node_id}] Trainer 初始化完成")

            # 4. 触发系统启动钩子
            if self.callbacks:
                self.sys_logger.debug(f"[{self.node_id}] 触发系统启动钩子")
                await self.callbacks.on_system_start(self)

            # 5. 运行主逻辑
            if self.trainer:
                return await self._run_as_trainer()
            else:
                return await self._run_as_learner()

        except Exception as e:
            self.sys_logger.exception(f"系统运行失败: {e}")
            raise

    async def stop(self):
        """停止系统"""
        self.sys_logger.info(f"Stopping FederatedSystem: {self.node_id}")

        if self.callbacks:
            await self.callbacks.on_system_stop(self)

        if self.tracker:
            self.tracker.close()
            self.sys_logger.info("Tracker closed")

        if self.node:
            await self.node.stop()

        self._connected = False
        self.sys_logger.debug(f"FederatedSystem stopped: {self.node_id}")

    # ========== 日志初始化 ==========

    def _setup_logging(self):
        """初始化日志系统"""
        log_config = self._config.logging or LogConfig()
        setup_logging(node_id=self.node_id, log_config=log_config)

    # ========== Node 创建 ==========

    async def _create_node(self):
        """创建通信节点"""
        # 从 NodeConfig 获取 NodeCommConfig
        # critical_peers 自动填充逻辑已在 get_comm_config() 中处理
        comm_config = self._config.get_comm_config()

        self.node = Node(comm_config)
        await self.node.initialize()

    # ========== 连接管理 ==========

    async def _ensure_connected(self):
        """确保连接已建立"""
        if self._connected:
            return

        self.sys_logger.debug(f"[{self.node_id}] 开始建立网络连接")

        # 1. 连接到远程节点（Learner 模式）
        if self._config.connect_to:
            retry_config = self._get_retry_config()
            self.sys_logger.debug(f"[{self.node_id}] 连接到远程节点: {self._config.connect_to}")

            for addr in self._config.connect_to:
                target_id, address = self._parse_address(addr)
                self.sys_logger.debug(f"[{self.node_id}] 正在连接到 {target_id}")
                await self.node.connect(target_id, address, retry_config=retry_config)
                self.sys_logger.debug(f"[{self.node_id}] 已成功连接到 {target_id}")

        # 2. 等待足够的 peer 连接（Trainer 模式）
        if self._config.listen and self._config.min_peers > 0:
            timeout = 120  # 默认超时
            self.sys_logger.info(
                f"[{self.node_id}] 等待 {self._config.min_peers} 个节点连接"
            )
            await self.node.wait_for_connections(
                min_peers=self._config.min_peers,
                timeout=timeout
            )
            connected_peers = self.node.get_connected_nodes()
            self.sys_logger.info(f"[{self.node_id}] 所有节点已连接: {connected_peers}")

        self._connected = True
        self.sys_logger.info(f"[{self.node_id}] 网络连接建立完成")

    def _get_retry_config(self) -> Dict[str, Any]:
        """获取连接重试配置"""
        return self._config.connection_retry.to_dict()

    def _extract_config_params(self) -> Dict[str, Any]:
        """
        提取配置参数用于记录到 MLflow
        
        委托给 NodeConfig.get_tracking_params()
        """
        return self._config.get_tracking_params()

    def _parse_address(self, addr: str) -> tuple:
        """解析地址字符串"""
        if "@" in addr:
            target_id, address = addr.split("@", 1)
        else:
            target_id = addr
            address = addr if ":" in addr else None
        return target_id, address

    # ========== 运行逻辑 ==========

    async def _run_as_trainer(self) -> Dict[str, Any]:
        """以 Trainer 角色运行"""
        self.sys_logger.info(f"[{self.node_id}] 开始运行 Trainer")
        result = await self.trainer.run()
        self.sys_logger.info(f"[{self.node_id}] Trainer 运行完成")
        return result

    async def _run_as_learner(self) -> Dict[str, Any]:
        """以 Learner 角色运行"""
        self.sys_logger.debug(f"[{self.node_id}] Learner 等待被调用")

        if self._shutdown_event:
            try:
                await self._shutdown_event.wait()
                self.sys_logger.debug(f"[{self.node_id}] 收到 shutdown 信号")

                if self.node:
                    self.sys_logger.info(f"[{self.node_id}] 正在停止 Node...")
                    await self.node.stop()

                if self.learner and hasattr(self.learner, 'on_shutdown'):
                    self.learner.on_shutdown()

                self.sys_logger.info(f"[{self.node_id}] Learner 已关闭")
            except asyncio.CancelledError:
                self.sys_logger.info(f"[{self.node_id}] Learner cancelled")
        else:
            try:
                await asyncio.Event().wait()
            except asyncio.CancelledError:
                self.sys_logger.info(f"[{self.node_id}] Learner stopped")

        return {
            "node_id": self.node_id,
            "role": "learner",
            "status": "shutdown",
        }

    # ========== 事件处理 ==========

    async def _on_peer_connect(self, peer_id: str, client_proxy):
        """连接事件处理器"""
        self.sys_logger.debug(f"Peer connected: {peer_id}")

        if self.learners:
            self.sys_logger.debug(f"Current learners count: {len(self.learners)}")

        if self.tracker:
            self.tracker.log_params({f"peer_{peer_id}": "connected"})

    async def _on_peer_disconnect(self, peer_id: str, reason: str = "unknown"):
        """断开连接事件处理器"""
        self.sys_logger.warning(f"Peer disconnected: {peer_id}, reason: {reason}")

        if self.learners:
            self.sys_logger.debug(f"Remaining learners count: {len(self.learners)}")

        if self.tracker:
            self.tracker.log_params({f"peer_{peer_id}": f"disconnected_{reason}"})

    # ========== 基础设施初始化 ==========

    async def _initialize_infrastructure(self):
        """初始化 Tracker 和 Callbacks"""
        self.sys_logger.debug("Initializing infrastructure...")

        # Tracker - 直接传递 TrackerConfig
        self.sys_logger.info(f"[DEBUG] self._config.tracker = {self._config.tracker}")
        if self._config.tracker:
            self.sys_logger.info(f"[DEBUG] tracker.enabled = {self._config.tracker.enabled}")
            self.sys_logger.info(f"[DEBUG] tracker.backends = {self._config.tracker.backends}")
            self.sys_logger.info(f"[DEBUG] exp_name = {self.exp_name}")
            
            try:
                self.tracker = CompositeTracker.from_config(
                    tracker_config=self._config.tracker,
                    node_id=self.node_id,
                    is_trainer=self._config.is_trainer(),
                    exp_name=self.exp_name,  # 传入实验名称
                )
                self.sys_logger.info(f"[DEBUG] Created tracker = {self.tracker}")
            except Exception as e:
                self.sys_logger.error(f"[DEBUG] Failed to create tracker: {e}")
                import traceback
                self.sys_logger.error(traceback.format_exc())
                self.tracker = None

            if self.tracker:
                # ===== 只有 Trainer 记录配置到 MLflow =====
                # Learner 的配置会通过同步机制记录到 Trainer 的 run 中
                if self._config.is_trainer():
                    params = self._extract_config_params()
                    self.tracker.log_params(params)
                    
                    self.tracker.set_tags({
                        "node_id": self.node_id,
                        "exp_name": self.exp_name,
                        "role": "trainer",
                        "framework": "oiafed",
                    })
                    
                    # 上传配置文件为 Artifact
                    if self._config.config_path:
                        import os
                        if os.path.exists(self._config.config_path):
                            try:
                                self.tracker.log_artifact(
                                    self._config.config_path, 
                                    artifact_path="configs"
                                )
                                self.sys_logger.debug(f"Uploaded config artifact: {self._config.config_path}")
                            except Exception as e:
                                self.sys_logger.warning(f"Failed to upload config artifact: {e}")

        # Callbacks
        self.callbacks = CallbackManager()
        for cb_config in self._config.get_callbacks():
            cb_args = cb_config.get_args()

            if cb_config.type == 'tracker_sync':
                if self.tracker:
                    cb_args['tracker'] = self.tracker
                # 传入 node_config，让 Learner 可以记录自己的配置
                cb_args['node_config'] = self._config

            cb = registry.create(
                namespace=f"federated.callback.{cb_config.type}",
                **cb_args
            )
            self.callbacks.add(cb)
            self.sys_logger.debug(f"Added Callback: {cb_config.type}")

        self.sys_logger.debug("Infrastructure initialized")

    # ========== 业务组件初始化 ==========

    async def _initialize_trainer(self):
        """初始化 Trainer"""
        self.sys_logger.info("Initializing Trainer...")

        # 1. 获取已连接的节点
        connected_peer_ids = self.node.get_connected_nodes()
        self.sys_logger.info(f"已连接的节点: {connected_peer_ids}")

        # 2. 创建 ProxyCollection
        self.learners = ProxyCollection(self.node, connected_peer_ids)
        self.sys_logger.info(f"Created ProxyCollection with {len(self.learners)} learners")

        # 3. 使用 Builder 构建依赖
        components = await self.builder.build(self._config)

        aggregator = components.get("aggregator")
        datasets_dict = components.get("datasets", {})
        model = components.get("model")  # ⭐ 获取模型（用于全局评估）

        # 向后兼容：取第一个训练数据集
        train_datasets = datasets_dict.get("train", [])
        dataset = train_datasets[0] if train_datasets else None

        # 4. 创建 Trainer
        trainer_config = self._config.get_trainer_config()
        self.sys_logger.info(f"正在创建 Trainer: {trainer_config.type}")
        self.sys_logger.info(f"[DEBUG] 创建 Trainer 时 self.tracker = {self.tracker}")
        self.sys_logger.info(f"[DEBUG] trainer_config.get_args() = {trainer_config.get_args()}")
        self.sys_logger.info(f"[DEBUG] model = {model}")  # ⭐ 打印模型信息

        # 构建 namespace
        trainer_type = trainer_config.type
        if "." not in trainer_type:
            namespace = f"trainer.{trainer_type}"
        else:
            namespace = trainer_type

        self.trainer = registry.create(
            namespace=namespace,
            config=trainer_config.get_args(),
            learners=self.learners,
            aggregator=aggregator,
            dataset=dataset,         # 向后兼容：第一个训练数据集
            datasets=datasets_dict,  # 新方式：完整的数据集字典
            model=model,             # ⭐ 传递模型（用于全局评估）
            tracker=self.tracker,
            callbacks=self.callbacks,
        )
        
        self.sys_logger.info(f"[DEBUG] 创建后 self.trainer._tracker = {getattr(self.trainer, '_tracker', 'NO ATTR')}")

        self.sys_logger.info(f"Trainer initialized: {trainer_config.type}")

    async def _initialize_learner(self):
        """初始化 Learner"""
        self.sys_logger.info("Initializing Learner...")

        # 使用 Builder 构建组件
        components = await self.builder.build(self._config)

        self.learner = components.get("learner")

        if self.learner:
            self.sys_logger.info(f"Created Learner: {type(self.learner).__name__}")

            if self.tracker and hasattr(self.learner, 'set_tracker'):
                self.learner.set_tracker(self.tracker)
            if self.callbacks and hasattr(self.learner, '_callbacks'):
                self.learner._callbacks = self.callbacks

            self.node.bind(self.learner, name="learner")

            self._shutdown_event = asyncio.Event()
            self.node.register("_fed_shutdown", self._handle_shutdown)
            self.sys_logger.info("Learner initialized and bound")
        else:
            self.sys_logger.warning("Learner not created")

    # ========== Shutdown 处理 ==========

    async def _handle_shutdown(self, payload: Dict[str, Any], ctx: Any) -> Dict[str, Any]:
        """处理 shutdown 信号"""
        reason = payload.get("reason", "unknown")
        self.sys_logger.debug(f"[{self.node_id}] 收到 shutdown 信号: {reason}")

        await self._stop_heartbeat_safely()

        if self._shutdown_event:
            self._shutdown_event.set()

        return {"status": "shutting_down", "reason": reason}

    async def _stop_heartbeat_safely(self):
        """安全停止心跳任务"""
        try:
            if not self.node or not hasattr(self.node, '_comm'):
                return

            comm = self.node._comm
            if not hasattr(comm, '_transport'):
                return

            transport = comm._transport

            if hasattr(transport, '_control_running') and transport._control_running:
                self.sys_logger.debug(f"[{self.node_id}] 停止心跳任务")
                transport._stop_control_thread()
                await asyncio.sleep(0.1)

        except Exception as e:
            self.sys_logger.warning(f"停止心跳任务时出错: {e}")

    # ========== 上下文管理器 ==========

    async def __aenter__(self):
        """异步上下文管理器入口"""
        await self.initialize()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """异步上下文管理器出口"""
        await self.stop()
        return False

    # ========== 辅助方法 ==========

    def get_experiment_info(self) -> Dict[str, Any]:
        """获取实验信息"""
        return {
            "node_id": self.node_id,
            "exp_name": self.exp_name,
            "run_name": self.run_name,
            "log_dir": self.log_dir,
            "is_trainer": self._config.is_trainer(),
            "is_learner": self._config.is_learner(),
            "initialized": self._initialized,
            "connected": self._connected,
        }

    # ========== 运行时信息同步 ==========

    def _generate_runtime_info(self) -> None:
        """
        [Trainer] 生成运行时信息
        
        在 Trainer 初始化时调用，生成统一的 run_name。
        这个 run_name 会通过 RPC 同步给所有 Learner。
        """
        # 使用配置中的 run_name，如果没有则生成
        if self._config.logging and self._config.logging.run_name:
            self._runtime_run_name = self._config.logging.run_name
        else:
            self._runtime_run_name = datetime.now().strftime("%Y%m%d_%H%M%S")
        
        # 更新配置中的 logging.run_name（确保日志系统使用统一的值）
        if self._config.logging:
            # LogConfig 是 dataclass，需要通过 object.__setattr__ 修改
            object.__setattr__(self._config.logging, 'run_name', self._runtime_run_name)

    def _register_runtime_sync_handler(self) -> None:
        """
        [Trainer] 注册运行时信息同步 RPC 方法
        
        Learner 连接后会调用此方法获取运行时信息。
        """
        async def handle_sync_runtime_info(
            payload: Dict[str, Any],
            ctx: Any,
        ) -> Dict[str, Any]:
            """处理运行时信息同步请求"""
            self.sys_logger.debug(f"收到运行时信息同步请求: from={ctx.source}")
            
            # 调用可覆盖的钩子方法
            result = self.get_sync_info()
            
            self.sys_logger.info(f"同步运行时信息到 {ctx.source}: {result}")
            return result
        
        self.node.register("_sync_runtime_info", handle_sync_runtime_info)
        self.sys_logger.debug("已注册运行时信息同步方法: _sync_runtime_info")

    def get_sync_info(self) -> Dict[str, Any]:
        """
        [Trainer] 获取要同步给 Learner 的运行时信息
        
        子类可覆盖此方法添加自定义同步信息。
        
        Returns:
            包含运行时信息的字典
            
        Example:
            class MySystem(FederatedSystem):
                def get_sync_info(self) -> Dict[str, Any]:
                    info = super().get_sync_info()
                    info["global_seed"] = 42
                    info["custom_config"] = {"key": "value"}
                    return info
        """
        info = {
            "exp_name": self.exp_name,
            "run_name": self._runtime_run_name,
        }
        
        # 如果有 MLflow Tracker，添加 run_id
        if self.tracker:
            for backend in getattr(self.tracker, 'trackers', []):
                if hasattr(backend, 'get_sync_info'):
                    info.update(backend.get_sync_info())
                    break
        
        return info

    async def _sync_runtime_info_from_trainer(self) -> None:
        """
        [Learner] 从 Trainer 同步运行时信息
        
        获取 Trainer 的 run_name 和 mlflow_run_id 等信息。
        """
        if self._runtime_synced:
            return
        
        # 获取 Trainer 节点 ID
        trainer_id = None
        if self._config.connect_to:
            for addr in self._config.connect_to:
                if '@' in addr:
                    trainer_id = addr.split('@')[0]
                    break
                else:
                    trainer_id = addr
                    break
        
        if not trainer_id:
            self.sys_logger.warning("无法确定 Trainer 节点 ID，跳过运行时信息同步")
            return
        
        try:
            self.sys_logger.info(f"从 Trainer ({trainer_id}) 同步运行时信息...")
            
            # 调用 Trainer 的同步方法
            result = await self.node.call(
                trainer_id,
                "_sync_runtime_info",
                {},
                timeout=10.0,
            )
            
            if result:
                self._runtime_run_name = result.get("run_name")
                self._runtime_mlflow_run_id = result.get("mlflow_run_id")
                
                self.sys_logger.info(
                    f"运行时信息同步成功: "
                    f"run_name={self._runtime_run_name}, "
                    f"mlflow_run_id={self._runtime_mlflow_run_id}"
                )
                
                # 调用可覆盖的钩子方法处理同步信息
                self.on_sync_info(result)
                
                self._runtime_synced = True
            else:
                self.sys_logger.warning("运行时信息同步返回空结果")
                
        except Exception as e:
            self.sys_logger.warning(f"运行时信息同步失败: {e}")
            # 不抛出异常，允许继续运行（使用本地配置）

    def on_sync_info(self, info: Dict[str, Any]) -> None:
        """
        [Learner] 处理从 Trainer 同步的运行时信息
        
        子类可覆盖此方法处理自定义同步信息。
        
        Args:
            info: 从 Trainer 同步的信息字典
            
        Example:
            class MySystem(FederatedSystem):
                def on_sync_info(self, info: Dict[str, Any]) -> None:
                    super().on_sync_info(info)  # 先处理默认逻辑
                    if "global_seed" in info:
                        torch.manual_seed(info["global_seed"])
        """
        # 默认处理：让 MLflow Tracker 加入 Trainer 的 run
        mlflow_run_id = info.get("mlflow_run_id")
        
        self.sys_logger.info(f"on_sync_info: mlflow_run_id={mlflow_run_id}, tracker={self.tracker}")
        
        if mlflow_run_id and self.tracker:
            backends = getattr(self.tracker, 'trackers', [])
            self.sys_logger.info(f"Found {len(backends)} tracker backends")
            
            for i, backend in enumerate(backends):
                backend_name = type(backend).__name__
                has_join_run = hasattr(backend, 'join_run')
                self.sys_logger.info(f"Backend[{i}]: {backend_name}, has_join_run={has_join_run}")
                
                if has_join_run:
                    backend.join_run(mlflow_run_id)
                    self.sys_logger.info(f"MLflow Tracker 已加入 run: {mlflow_run_id}")
                    break
            
            # ===== 记录 Learner 配置参数 =====
            self._log_learner_config_to_tracker()
            
        elif not mlflow_run_id:
            self.sys_logger.warning("on_sync_info: mlflow_run_id is None")
        elif not self.tracker:
            self.sys_logger.warning("on_sync_info: self.tracker is None")
    
    def _log_learner_config_to_tracker(self) -> None:
        """
        记录 Learner 配置参数到 Tracker
        
        在 join_run 之后调用，确保参数记录到正确的 run 中
        """
        if not self.tracker or not self._config:
            return
        
        try:
            # 获取配置参数
            params = self._config.get_tracking_params()
            print(f"[system] _log_learner_config_to_tracker: Got {len(params)} params")
            
            # 记录参数
            if hasattr(self.tracker, "log_params"):
                self.tracker.log_params(params)
                print(f"[system] Logged {len(params)} params for {self.node_id}")
            
            # 设置标签
            if hasattr(self.tracker, "set_tags"):
                self.tracker.set_tags({
                    f"peer_{self.node_id}": "connected",
                })
                
        except Exception as e:
            print(f"[system] ERROR: Failed to log learner config: {e}")
            import traceback
            traceback.print_exc()

    def __repr__(self) -> str:
        """字符串表示"""
        role = "trainer" if self._config.is_trainer() else "learner"
        return f"FederatedSystem(node_id={self.node_id}, role={role}, exp={self.exp_name})"


"""
FedProto (Federated Prototypical Learning) 学习器实现

从 methods/learners/fl/fedproto.py 迁移到新架构

论文：FedProto: Federated Prototype Learning across Heterogeneous Clients
作者：Yue Tan et al.
发表：AAAI 2022

FedProto的核心思想：
- 使用原型（prototypes）来表示每个类别的特征中心
- 客户端之间共享类别原型而不是模型参数
- 使用原型进行知识蒸馏，提高泛化性能
"""
from typing import Dict, Any, Optional
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import DataLoader

from ....core.learner import Learner
from ....core.types import StepMetrics, EpochMetrics, EvalResult
from ....registry import learner


@learner('fl.fedproto', description='FedProto: Federated Prototypical Learning')
class FedProtoLearner(Learner):
    """FedProto学习器 - 使用原型学习的联邦学习

    配置示例:
    {
        "learner": {
            "name": "fl.fedproto",
            "learning_rate": 0.01,
            "batch_size": 128,
            "local_epochs": 5,
            "optimizer": "SGD",
            "momentum": 0.9,
            "num_classes": 10,
            "lambda_proto": 1.0,  # 原型损失的权重
            "temperature": 0.5  # 原型对比的温度参数
        }
    }
    """

    def __init__(
        self,
        model,
        datasets,
        tracker=None,
        callbacks=None,
        config=None,
        node_id=None
    ):
        super().__init__(model, None, tracker, callbacks, config, node_id)

        self._datasets = datasets or {}

        # 从配置提取参数
        self.learning_rate = self._config.get('learning_rate', 0.01)
        self.batch_size = self._config.get('batch_size', 32)
        self.local_epochs = self._config.get('local_epochs', 5)
        self.momentum = self._config.get('momentum', 0.9)
        self.optimizer_type = self._config.get('optimizer', 'SGD').upper()

        # FedProto特有参数
        self.num_classes = self._config.get('num_classes', 10)
        self.lambda_proto = self._config.get('lambda_proto', 1.0)
        self.temperature = self._config.get('temperature', 0.5)

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        # 组件占位符
        self._optimizer = None
        self._criterion = None
        self._train_loader = None
        self._test_loader = None

        # 原型相关
        self.local_prototypes = None  # 本地原型
        self.global_prototypes = None  # 全局原型
        self.round_number = 0

        self.logger.info(
            f"FedProtoLearner {node_id} 初始化完成 "
            f"(lr={self.learning_rate}, num_classes={self.num_classes}, lambda_proto={self.lambda_proto})"
        )

    async def setup(self, config: Dict) -> None:
        """初始化训练环境"""
        # 获取PyTorch模型
        self.model.to(self.device)

        # 创建优化器
        if self.optimizer_type == 'SGD':
            self._optimizer = optim.SGD(
                self.model.parameters(),
                lr=self.learning_rate,
                momentum=self.momentum
            )
        elif self.optimizer_type == 'ADAM':
            self._optimizer = optim.Adam(
                self.model.parameters(),
                lr=self.learning_rate
            )
        else:
            raise ValueError(f"不支持的优化器类型: {self.optimizer_type}")

        # 创建损失函数
        loss_name = self._config.get('loss', 'CrossEntropyLoss')
        if loss_name == 'CrossEntropyLoss':
            self._criterion = nn.CrossEntropyLoss()
        elif loss_name == 'MSELoss':
            self._criterion = nn.MSELoss()
        else:
            raise ValueError(f"不支持的损失函数: {loss_name}")

        # 创建数据加载器
        train_datasets = self._datasets.get("train", [])
        if train_datasets:
            self._train_loader = DataLoader(
                train_datasets[0],
                batch_size=self.batch_size,
                shuffle=True
            )

        test_datasets = self._datasets.get("test", [])
        if test_datasets:
            self._test_loader = DataLoader(
                test_datasets[0],
                batch_size=self.batch_size,
                shuffle=False
            )

        self.logger.info(f"Setup完成: train_samples={len(train_datasets[0]) if train_datasets else 0}")

    def get_features(self, model, data):
        """获取模型的特征表示（倒数第二层的输出）"""
        last_fc_candidates = ['fc2', 'fc', 'classifier', 'head']

        last_fc_name = None
        for name in last_fc_candidates:
            if hasattr(model, name):
                last_fc_name = name
                break

        if last_fc_name:
            original_fc = getattr(model, last_fc_name)
            setattr(model, last_fc_name, nn.Identity())
            features = model(data)
            setattr(model, last_fc_name, original_fc)
            return features
        else:
            # 使用hook方式获取倒数第二层的输出
            features = []

            def hook_fn(module, input, output):
                features.append(output)

            modules = list(model.children())
            if len(modules) >= 2:
                handle = modules[-2].register_forward_hook(hook_fn)
                _ = model(data)
                handle.remove()

                if features:
                    feat = features[0]
                    if len(feat.shape) > 2:
                        feat = torch.flatten(feat, 1)
                    return feat

            self.logger.warning(
                f"无法提取特征，使用模型完整输出。"
                f"建议：确保模型有 'fc', 'fc2', 'classifier' 或 'head' 属性"
            )
            output = model(data)
            if len(output.shape) > 2:
                output = torch.flatten(output, 1)
            return output

    def compute_prototypes(self):
        """计算本地原型"""
        self.model.eval()

        # 初始化原型存储
        proto_features = {i: [] for i in range(self.num_classes)}

        with torch.no_grad():
            for data, target in self._train_loader:
                data = data.to(self.device)
                features = self.get_features(self.model, data)

                # 按类别收集特征
                for i in range(len(target)):
                    label = target[i].item()
                    proto_features[label].append(features[i].cpu())

        # 计算每个类别的原型（平均特征）
        prototypes = {}
        for label in range(self.num_classes):
            if len(proto_features[label]) > 0:
                prototypes[label] = torch.stack(proto_features[label]).mean(dim=0)
            else:
                # 如果该类别没有样本，使用零向量
                feature_dim = next(iter(proto_features.values()))[0].shape[0] if any(proto_features.values()) else 128
                prototypes[label] = torch.zeros(feature_dim)

        self.local_prototypes = prototypes
        self.logger.info(
            f"[{self._node_id}] Computed prototypes for "
            f"{len([p for p in prototypes.values() if p.sum() != 0])} classes"
        )

        return prototypes

    def prototype_loss(self, features, targets):
        """
        计算原型损失

        Args:
            features: 特征向量 (batch_size, feature_dim)
            targets: 真实标签 (batch_size,)

        Returns:
            原型损失值
        """
        if self.global_prototypes is None:
            return torch.tensor(0.0, device=self.device)

        batch_loss = 0.0
        valid_samples = 0

        for i in range(len(targets)):
            label = targets[i].item()
            feature = features[i]

            # 获取全局原型
            if label not in self.global_prototypes:
                continue

            proto = self.global_prototypes[label].to(self.device)

            # 计算与原型的距离（使用余弦相似度）
            feature_norm = F.normalize(feature.unsqueeze(0), dim=1)
            proto_norm = F.normalize(proto.unsqueeze(0), dim=1)

            similarity = torch.sum(feature_norm * proto_norm)

            # 负相似度作为损失（希望特征接近原型）
            batch_loss += -similarity
            valid_samples += 1

        if valid_samples > 0:
            return batch_loss / valid_samples
        else:
            return torch.tensor(0.0, device=self.device)

    async def train_step(self, batch: Any, batch_idx: int) -> StepMetrics:
        """
        单批次训练 - FedProto 训练步骤
        
        包含交叉熵损失和原型损失。
        """
        data, target = batch
        data, target = data.to(self.device), target.to(self.device)

        self._optimizer.zero_grad()

        # 前向传播
        output = self.model(data)
        ce_loss = self._criterion(output, target)

        # 计算原型损失
        proto_loss = torch.tensor(0.0, device=self.device)
        if self.global_prototypes is not None and self.round_number > 1:
            features = self.get_features(self.model, data)
            proto_loss = self.prototype_loss(features, target)

        # 总损失
        loss = ce_loss + self.lambda_proto * proto_loss

        loss.backward()
        self._optimizer.step()

        # 计算准确率
        pred = output.argmax(dim=1, keepdim=True)
        correct = pred.eq(target.view_as(pred)).sum().item()
        accuracy = correct / data.size(0)

        return StepMetrics(
            loss=loss.item(),
            batch_size=data.size(0),
            metrics={
                'accuracy': accuracy,
                'ce_loss': ce_loss.item(),
                'proto_loss': proto_loss.item()
            }
        )

    async def train_epoch(self, epoch_idx: int) -> EpochMetrics:
        """单轮训练 - FedProto训练循环"""
        self.model.train()

        total_loss = 0.0
        total_ce_loss = 0.0
        total_proto_loss = 0.0
        total_correct = 0
        total_samples = 0

        for data, target in self._train_loader:
            data, target = data.to(self.device), target.to(self.device)

            self._optimizer.zero_grad()

            # 前向传播
            output = self.model(data)
            ce_loss = self._criterion(output, target)

            # 计算原型损失
            proto_loss = torch.tensor(0.0, device=self.device)
            if self.global_prototypes is not None and self.round_number > 1:
                features = self.get_features(self.model, data)
                proto_loss = self.prototype_loss(features, target)

            # 总损失
            loss = ce_loss + self.lambda_proto * proto_loss

            loss.backward()
            self._optimizer.step()

            total_loss += loss.item() * data.size(0)
            total_ce_loss += ce_loss.item() * data.size(0)
            total_proto_loss += proto_loss.item() * data.size(0)

            pred = output.argmax(dim=1, keepdim=True)
            total_correct += pred.eq(target.view_as(pred)).sum().item()
            total_samples += data.size(0)

        avg_loss = total_loss / total_samples
        avg_accuracy = total_correct / total_samples

        self.logger.info(
            f"[{self._node_id}] Epoch {epoch_idx}: "
            f"Loss={avg_loss:.4f} (CE={total_ce_loss/total_samples:.4f}, "
            f"Proto={total_proto_loss/total_samples:.4f}), Acc={avg_accuracy:.4f}"
        )

        epoch_metrics = EpochMetrics(
            epoch=epoch_idx,
            avg_loss=avg_loss,
            total_samples=total_samples,
            metrics={
                'accuracy': avg_accuracy,
                'ce_loss': total_ce_loss / total_samples,
                'proto_loss': total_proto_loss / total_samples
            }
        )

        # 增加全局 epoch 计数器（用于 MLflow step）
        self._global_epoch_counter += 1

        # 触发 epoch 结束回调
        if self._callbacks:
            await self._callbacks.on_epoch_end(self, epoch_idx, epoch_metrics)

        return epoch_metrics

    async def fit(self, config: Optional[Dict] = None):
        """训练模型 - 重写以计算原型"""
        # 增加轮次计数
        self.round_number += 1

        # 调用父类的fit方法
        await super().fit(config)

        # 训练后计算本地原型
        self.compute_prototypes()

    async def evaluate(self, config: Optional[Dict] = None) -> EvalResult:
        """评估模型"""
        self.model.eval()

        loader = self._test_loader if self._test_loader else self._train_loader

        total_loss = 0.0
        total_correct = 0
        total_samples = 0

        with torch.no_grad():
            for data, target in loader:
                data, target = data.to(self.device), target.to(self.device)
                output = self.model(data)
                loss = self._criterion(output, target)

                total_loss += loss.item() * data.size(0)
                pred = output.argmax(dim=1, keepdim=True)
                total_correct += pred.eq(target.view_as(pred)).sum().item()
                total_samples += data.size(0)

        return EvalResult(
            num_samples=total_samples,
            metrics={
                'loss': total_loss / total_samples,
                'accuracy': total_correct / total_samples
            }
        )

    async def get_weights(self) -> Dict[str, Any]:
        """获取模型权重和本地原型"""
        weights = {name: param.data.clone() for name, param in self.model.state_dict().items()}

        # 添加原型信息
        if self.local_prototypes is not None:
            weights['_prototypes'] = self.local_prototypes

        return weights

    async def set_weights(self, weights: Dict[str, Any]):
        """设置模型权重和全局原型"""
        # 提取原型信息
        if '_prototypes' in weights:
            self.global_prototypes = weights.pop('_prototypes')
            self.logger.info(
                f"[{self._node_id}] Updated global prototypes for "
                f"{len(self.global_prototypes)} classes"
            )

        # 更新模型权重
        torch_weights = {}
        for k, v in weights.items():
            if torch.is_tensor(v):
                torch_weights[k] = v
            else:
                torch_weights[k] = torch.from_numpy(v)

        self.model.load_state_dict(torch_weights, strict=False)


"""
MLP 模型 (表格数据)

包含适用于 Adult、FCUBE 等表格数据集的 MLP 模型。

注册名:
- paper_mlp: 通用 MLP (TPAMI 2025)
- adult_mlp: Adult 数据集专用
- fcube_mlp: FCUBE 数据集专用
"""

import torch
import torch.nn as nn

from ...registry.decorators import model


# ==================== 通用 MLP (TPAMI 2025) ====================

@model(
    name='paper_mlp',
    description='论文标准 MLP (TPAMI 2025) - 适配表格数据',
    task='classification',
    version='1.0'
)
class PaperMLP(nn.Module):
    """
    论文标准 MLP 模型
    
    符合 TPAMI 2025 论文 Section VI 标准架构。
    
    架构:
    "For the tabular datasets, we employ a standard MLP with
    three hidden layers (32, 16, and 8 units)."
    
    - FC1: input_dim → 32, ReLU
    - FC2: 32 → 16, ReLU
    - FC3: 16 → 8, ReLU
    - FC4: 8 → num_classes
    
    支持的数据集:
    - Adult: 99 特征, 2 分类
    - FCUBE: 3 特征, 2 分类
    """

    def __init__(self, input_dim: int, num_classes: int = 2):
        """
        Args:
            input_dim: 输入特征维度
            num_classes: 分类数量
        """
        super().__init__()

        self.fc1 = nn.Linear(input_dim, 32)
        self.fc2 = nn.Linear(32, 16)
        self.fc3 = nn.Linear(16, 8)
        self.fc4 = nn.Linear(8, num_classes)

    def forward(self, x):
        # 确保输入是 2D
        if x.dim() == 1:
            x = x.unsqueeze(0)

        x = torch.relu(self.fc1(x))
        x = torch.relu(self.fc2(x))
        x = torch.relu(self.fc3(x))
        x = self.fc4(x)
        return x


# ==================== Adult 数据集专用 ====================

@model(
    name='adult_mlp',
    description='Adult 数据集 MLP (99 特征, 2 分类)',
    task='classification',
    version='1.0',
    input_shape=(99,),
    output_shape=(2,)
)
class AdultMLP(PaperMLP):
    """
    Adult 数据集专用 MLP
    
    输入: 99 维特征向量
    输出: 2 分类 (收入 >50K / <=50K)
    """

    def __init__(self, num_classes: int = 2):
        super().__init__(input_dim=99, num_classes=num_classes)


# ==================== FCUBE 数据集专用 ====================

@model(
    name='fcube_mlp',
    description='FCUBE 数据集 MLP (3 特征, 2 分类)',
    task='classification',
    version='1.0',
    input_shape=(3,),
    output_shape=(2,)
)
class FCUBEMLP(PaperMLP):
    """
    FCUBE 数据集专用 MLP
    
    输入: 3 维特征向量
    输出: 2 分类
    """

    def __init__(self, num_classes: int = 2):
        super().__init__(input_dim=3, num_classes=num_classes)


# ==================== 导出 ====================

__all__ = [
    "PaperMLP",
    "AdultMLP",
    "FCUBEMLP",
]



"""
FedNova 聚合器

从 methods/aggregators/fednova.py 迁移到 src/

实现 FedNova (Federated Optimization in Heterogeneous Networks) 聚合算法。
通过标准化客户端更新来处理客户端异构性问题。

论文: Tackling the Objective Inconsistency Problem in Heterogeneous Federated Optimization
作者: Jianyu Wang et al.
发表: NeurIPS 2020

算法特点:
1. 标准化不同客户端的本地更新步数
2. 更好地处理系统异构性
3. 改善收敛速度和稳定性
"""

import torch
from typing import List, Dict, Any
from loguru import logger

from ...core.aggregator import Aggregator
from ...core.types import ClientUpdate
from ...registry import aggregator


@aggregator(
    name='fednova',
    description='FedNova标准化联邦聚合器',
    version='1.0'
)
class FedNovaAggregator(Aggregator):
    """FedNova 聚合器实现"""

    def __init__(self, **kwargs):
        """初始化FedNova聚合器"""
        device = kwargs.get("device", "auto")
        if device == "auto":
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self.device = device

        self.round_count = 0

        logger.info("✅ FedNova聚合器初始化完成")

    def aggregate(self, updates: List[ClientUpdate], global_model=None) -> Dict[str, torch.Tensor]:
        """
        执行FedNova聚合

        Args:
            updates: 客户端更新列表 (List[ClientUpdate])
            global_model: 全局模型 (可选)

        Returns:
            聚合后的模型权重字典
        """
        if not updates:
            raise ValueError("没有客户端更新可聚合")

        self.round_count += 1

        # 计算有效批次数进行标准化
        effective_batches = []
        for update in updates:
            local_epochs = 1
            if hasattr(update, 'metadata') and update.metadata:
                local_epochs = update.metadata.get("local_epochs", 1)
            effective_batches.append(local_epochs * update.num_samples)

        total_effective = sum(effective_batches)

        # 标准化权重
        weights = [eff / total_effective for eff in effective_batches]

        # 聚合
        aggregated_weights = {}
        first_weights = updates[0].weights

        for param_name in first_weights:
            # 创建float类型的零张量，避免Long类型转换错误
            aggregated_weights[param_name] = torch.zeros_like(
                first_weights[param_name], dtype=torch.float32, device=self.device
            )

            for i, update in enumerate(updates):
                param_value = update.weights[param_name].to(self.device)
                # 如果原始参数是Long类型，将其转换为float以便聚合
                if param_value.dtype in [torch.long, torch.int, torch.int32, torch.int64]:
                    param_value = param_value.float()
                aggregated_weights[param_name] += weights[i] * param_value

            # 如果原始参数是Long类型，将聚合结果转换回Long
            if first_weights[param_name].dtype in [torch.long, torch.int, torch.int32, torch.int64]:
                aggregated_weights[param_name] = aggregated_weights[param_name].long()

        logger.debug(f"✅ FedNova聚合完成 - 轮次 {self.round_count}")

        return aggregated_weights

    def get_stats(self) -> Dict[str, Any]:
        """获取聚合器统计信息"""
        return {
            "algorithm": "FedNova",
            "rounds": self.round_count
        }



"""
传输层模块
"""

from .base import Transport
from .memory import MemoryTransport
from .grpc_transport import GrpcTransport
from .factory import create_transport

__all__ = [
    "Transport",
    "MemoryTransport",
    "GrpcTransport",
    "create_transport",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
OiaFed - One Framework for All Federation

统一的联邦学习框架，支持所有联邦场景
"""

from setuptools import setup, find_packages
from pathlib import Path

# 读取 README
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

# 读取版本号
def get_version():
    version_file = this_directory / "oiafed" / "__init__.py"
    for line in version_file.read_text().splitlines():
        if line.startswith("__version__"):
            return line.split('"')[1]
    return "0.1.0"

# 核心依赖
INSTALL_REQUIRES = [
    # 深度学习框架
    "torch>=1.12.0",
    "torchvision>=0.13.0",
    "torchaudio>=0.12.0",
    
    # gRPC 通信（必需）
    "grpcio>=1.50.0",
    "grpcio-tools>=1.50.0",
    "protobuf>=3.20.0",
    
    # 配置和日志
    "pyyaml>=6.0",
    "omegaconf>=2.2.0",
    "loguru>=0.6.0",
    "toml>=0.10.0",
    
    # 数据处理
    "numpy>=1.21.0",
    "pandas>=1.4.0",
    "scikit-learn>=1.0.0",
    
    # 可视化
    "matplotlib>=3.5.0",
    "tqdm>=4.60.0",
    
    # 网络和异步
    "aiohttp>=3.8.0",
    
    # 工具
    "psutil>=5.9.0",
    "openpyxl>=3.0.0",
]

# 可选依赖
EXTRAS_REQUIRE = {
    # MLflow 实验追踪
    "mlflow": [
        "mlflow>=2.0.0",
    ],
    # 开发依赖
    "dev": [
        "pytest>=7.0.0",
        "pytest-asyncio>=0.20.0",
        "pytest-html>=3.2.0",
        "pytest-cov>=4.0.0",
        "black>=22.0.0",
        "isort>=5.10.0",
        "mypy>=0.990",
        "flake8>=5.0.0",
    ],
    # 文档
    "docs": [
        "sphinx>=5.0.0",
        "sphinx-rtd-theme>=1.0.0",
        "myst-parser>=0.18.0",
    ],
}

# all = mlflow + dev
EXTRAS_REQUIRE["all"] = (
    EXTRAS_REQUIRE["mlflow"] + 
    EXTRAS_REQUIRE["dev"] + 
    EXTRAS_REQUIRE["docs"]
)

setup(
    name="oiafed",
    version=get_version(),
    author="OiaFed Team",
    author_email="contact@oiafed.com",
    description="OiaFed: One Framework for All Federation - A unified federated learning framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/oiafed/oiafed",
    project_urls={
        "Homepage": "https://oiafed.cn",
        "Documentation": "https://docs.oiafed.cn",
        "Repository": "https://github.com/oiafed/oiafed",
        "Issues": "https://github.com/oiafed/oiafed/issues",
        "Changelog": "https://github.com/oiafed/oiafed/blob/main/CHANGELOG.md",
    },
    license="MIT",
    
    # 包配置
    packages=find_packages(exclude=["tests", "tests.*", "examples", "docs"]),
    py_modules=[],
    include_package_data=True,
    
    # 依赖
    python_requires=">=3.10",
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    
    # 入口点
    entry_points={
        "console_scripts": [
            "oiafed=oiafed.cli:main",
        ],
    },
    
    # 分类
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: System :: Distributed Computing",
    ],
    keywords=[
        "federated-learning",
        "machine-learning", 
        "deep-learning",
        "distributed-computing",
        "continual-learning",
        "personalized-federated-learning",
        "privacy-preserving",
    ],
)


"""
内置 Callback 实现

提供常用的 Callback：
- ModelCheckpoint: 模型检查点
- EarlyStopping: 早停
"""

import os
from typing import Dict, Any
from .base import Callback
from ..registry import register
from ..infra.logging import get_module_logger

logger = get_module_logger(__name__)


@register("federated.callback.model_checkpoint")
class ModelCheckpoint(Callback):
    """
    模型检查点 Callback

    定期保存模型
    """

    def __init__(self, save_dir: str = "./checkpoints", save_freq: int = 10):
        """
        Args:
            save_dir: 保存目录
            save_freq: 保存频率（每 N 轮保存一次）
        """
        self.save_dir = save_dir
        self.save_freq = save_freq

        # 创建目录
        os.makedirs(save_dir, exist_ok=True)
        # 使用基础logger输出初始化信息（在日志系统初始化之前）
        logger.info(f"ModelCheckpoint initialized: save_dir={save_dir}, save_freq={save_freq}")

    async def on_round_end(self, trainer: Any, round_num: int, context: Dict[str, Any] = None):
        """每 N 轮保存一次模型"""
        if (round_num + 1) % self.save_freq == 0:
            # 获取模型
            model = getattr(trainer, "model", None)
            # 获取trainer的logger
            train_logger = getattr(trainer, "_train_logger", logger)

            if model and hasattr(model, "save"):
                path = os.path.join(self.save_dir, f"model_round_{round_num + 1}.pt")
                try:
                    model.save(path)
                    train_logger.info(f"Model saved: {path}")
                except Exception as e:
                    train_logger.error(f"Failed to save model: {e}")
            else:
                train_logger.debug(f"Trainer has no saveable model at round {round_num + 1}")


@register("federated.callback.logging")
class LoggingCallback(Callback):
    """
    训练日志 Callback

    在训练过程中打印日志信息
    """

    def __init__(
        self,
        log_epoch: bool = True,
        log_round: bool = True,
        log_fit: bool = False,
    ):
        """
        Args:
            log_epoch: 是否记录 epoch 级别日志
            log_round: 是否记录 round 级别日志
            log_fit: 是否记录 fit 开始/结束日志
        """
        self.log_epoch = log_epoch
        self.log_round = log_round
        self.log_fit = log_fit

        logger.debug(
            f"LoggingCallback initialized: log_epoch={log_epoch}, "
            f"log_round={log_round}, log_fit={log_fit}"
        )

    async def on_fit_start(self, learner: Any, config: Dict[str, Any] = None):
        """Fit 开始时打印日志"""
        if not self.log_fit:
            return

        node_id = getattr(learner, "_node_id", "unknown")
        epochs = config.get("epochs", "?") if config else "?"
        # 获取learner的logger
        train_logger = getattr(learner, "_train_logger", logger)
        train_logger.debug(f"[{node_id}] Starting fit: epochs={epochs}")

    async def on_fit_end(self, learner: Any, result: Any = None):
        """Fit 结束时打印日志"""
        if not self.log_fit:
            return

        node_id = getattr(learner, "_node_id", "unknown")
        # 获取learner的logger
        train_logger = getattr(learner, "_train_logger", logger)
        train_logger.debug(f"[{node_id}] Fit completed")

    async def on_epoch_end(self, learner: Any, epoch: int, metrics: Any = None):
        """Epoch 结束时打印日志"""
        if not self.log_epoch:
            return

        node_id = getattr(learner, "_node_id", "unknown")
        # 获取learner的logger
        train_logger = getattr(learner, "_train_logger", logger)

        if metrics:
            # 从 EpochMetrics 中提取信息
            avg_loss = getattr(metrics, "avg_loss", "N/A")
            total_samples = getattr(metrics, "total_samples", 0)
            extra_metrics = getattr(metrics, "metrics", {})

            # 构建日志消息
            msg_parts = [f"[{node_id}] Epoch {epoch}"]
            msg_parts.append(f"loss={avg_loss:.4f}" if isinstance(avg_loss, float) else f"loss={avg_loss}")
            msg_parts.append(f"samples={total_samples}")

            # 添加额外指标（如 accuracy）
            for key, value in extra_metrics.items():
                if isinstance(value, float):
                    msg_parts.append(f"{key}={value:.4f}")
                else:
                    msg_parts.append(f"{key}={value}")

            train_logger.info(", ".join(msg_parts))
        else:
            train_logger.info(f"[{node_id}] Epoch {epoch} completed")

    async def on_round_end(self, trainer: Any, round_num: int, context: Dict[str, Any] = None):
        """Round 结束时打印日志"""
        if not self.log_round:
            return

        # 获取trainer的logger
        train_logger = getattr(trainer, "_train_logger", logger)

        if context and "metrics" in context:
            metrics = context["metrics"]

            # 从 RoundMetrics 中提取信息
            if hasattr(metrics, "num_clients"):
                num_clients = metrics.num_clients
                total_samples = metrics.total_samples
                metrics_dict = metrics.metrics
            else:
                # 兼容字典格式
                num_clients = metrics.get("num_clients", 0)
                total_samples = metrics.get("total_samples", 0)
                metrics_dict = metrics.get("metrics", {})

            # 构建日志消息
            msg_parts = [f"Round {round_num}"]
            msg_parts.append(f"clients={num_clients}")
            msg_parts.append(f"samples={total_samples}")

            # 添加指标
            for key, value in metrics_dict.items():
                if isinstance(value, float):
                    msg_parts.append(f"{key}={value:.4f}")
                else:
                    msg_parts.append(f"{key}={value}")

            train_logger.debug(", ".join(msg_parts))
        else:
            train_logger.debug(f"Round {round_num} completed")


@register("federated.callback.early_stopping")
class EarlyStopping(Callback):
    """
    早停 Callback

    如果指标不再改善，提前停止训练
    """

    def __init__(
        self,
        monitor: str = "loss",
        patience: int = 5,
        min_delta: float = 0.001,
        mode: str = "min"
    ):
        """
        Args:
            monitor: 监控的指标名称
            patience: 容忍轮次（连续多少轮不改善就停止）
            min_delta: 最小改善量
            mode: 'min' 或 'max'（指标越小越好还是越大越好）
        """
        self.monitor = monitor
        self.patience = patience
        self.min_delta = min_delta
        self.mode = mode

        self.best_value = None
        self.wait = 0
        self.stopped = False

        logger.debug(
            f"EarlyStopping initialized: monitor={monitor}, patience={patience}, "
            f"min_delta={min_delta}, mode={mode}"
        )

    async def on_round_end(self, trainer: Any, round_num: int, context: Dict[str, Any] = None):
        """检查是否需要早停"""
        if not context or self.monitor not in context:
            return

        current_value = context[self.monitor]
        # 获取trainer的logger
        train_logger = getattr(trainer, "_train_logger", logger)

        # 第一次
        if self.best_value is None:
            self.best_value = current_value
            train_logger.debug(f"EarlyStopping: initial {self.monitor}={current_value:.4f}")
            return

        # 检查是否改善
        improved = False
        if self.mode == "min":
            if current_value < self.best_value - self.min_delta:
                improved = True
        else:  # mode == "max"
            if current_value > self.best_value + self.min_delta:
                improved = True

        if improved:
            self.best_value = current_value
            self.wait = 0
            train_logger.debug(
                f"EarlyStopping: {self.monitor} improved to {current_value:.4f}, "
                f"reset wait counter"
            )
        else:
            self.wait += 1
            train_logger.debug(
                f"EarlyStopping: {self.monitor}={current_value:.4f} no improvement, "
                f"wait={self.wait}/{self.patience}"
            )

        # 触发早停
        if self.wait >= self.patience:
            train_logger.debug(
                f"Early stopping triggered at round {round_num + 1}: "
                f"{self.monitor} has not improved for {self.patience} rounds"
            )
            self.stopped = True

            # 通知 Trainer 停止（需要 Trainer 支持）
            if hasattr(trainer, "stop_training"):
                trainer.stop_training = True
                train_logger.debug("Set trainer.stop_training = True")


@register("federated.callback.mlflow")
class MLflowCallback(Callback):
    """
    MLflow Callback

    记录训练指标到 MLflow：
    - Learner: 每个 epoch 结束时记录指标
    - Learner: 评估结束时记录评估指标
    - Trainer: 每个 round 结束时记录指标
    """

    def __init__(self, log_epoch: bool = True, log_round: bool = True, log_eval: bool = True):
        """
        Args:
            log_epoch: 是否记录 epoch 级别指标（Learner）
            log_round: 是否记录 round 级别指标（Trainer）
            log_eval: 是否记录评估指标（Learner）
        """
        self.log_epoch = log_epoch
        self.log_round = log_round
        self.log_eval = log_eval

        logger.debug(
            f"MLflowCallback initialized: log_epoch={log_epoch}, log_round={log_round}, log_eval={log_eval}"
        )

    async def on_epoch_end(self, learner: Any, epoch: int, metrics: Any = None):
        """Learner Epoch 结束时记录指标到 MLflow"""
        if not self.log_epoch:
            return

        tracker = getattr(learner, "_tracker", None)
        if not tracker:
            return

        if metrics:
            # 从 EpochMetrics 中提取信息
            avg_loss = getattr(metrics, "avg_loss", None)
            total_samples = getattr(metrics, "total_samples", 0)
            extra_metrics = getattr(metrics, "metrics", {})

            # 构建指标字典
            metrics_dict = {}
            if isinstance(avg_loss, float):
                metrics_dict["loss"] = avg_loss

            # 添加额外指标（如 accuracy）
            metrics_dict.update(extra_metrics)

            # 使用全局 epoch 计数器作为 step（累积的，跨多次 fit 调用）
            global_epoch = getattr(learner, "_global_epoch_counter", epoch)
            tracker.log_metrics(metrics_dict, step=global_epoch)

    async def on_round_end(self, trainer: Any, round_num: int, context: Dict[str, Any] = None):
        """Trainer Round 结束时记录指标到 MLflow"""

        if not self.log_round:
            return

        tracker = getattr(trainer, "_tracker", None)

        if not tracker:
            return


        if context and "metrics" in context:
            metrics = context["metrics"]

            # 从 RoundMetrics 中提取信息
            if hasattr(metrics, "metrics"):
                # RoundMetrics 对象
                metrics_dict = dict(metrics.metrics)
            elif isinstance(metrics, dict):
                # 字典格式（兼容）
                metrics_dict = metrics.get("metrics", {})
            else:
                return


            # 记录到 tracker，使用 round_num 作为 step
            tracker.log_metrics(metrics_dict, step=round_num)

    async def on_evaluate_end(self, learner: Any, result: Any = None, context: Dict[str, Any] = None):
        """Learner 评估结束时记录评估指标到 MLflow"""
        if not self.log_eval:
            return

        tracker = getattr(learner, "_tracker", None)
        if not tracker:
            return

        if result and hasattr(result, "metrics"):
            # 从 EvalResult 中提取评估指标
            eval_metrics = result.metrics

            # 添加评估类型前缀（如果有上下文信息）
            if context:
                eval_type = context.get("type", "eval")
                # 如果是 post_training 评估，使用 eval_ 前缀
                # 否则保持原有的 post_train_ 前缀（这个会在 learner.py 中添加）
                if eval_type == "post_training":
                    metrics_to_log = {f"eval_{k}": v for k, v in eval_metrics.items()}
                else:
                    metrics_to_log = eval_metrics
            else:
                metrics_to_log = {f"eval_{k}": v for k, v in eval_metrics.items()}

            # 使用全局 epoch 计数器作为 step（累积的，跨多次 fit 调用）
            global_epoch = getattr(learner, "_global_epoch_counter", None)
            tracker.log_metrics(metrics_to_log, step=global_epoch)



"""
拦截器抽象基类和拦截器链
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Optional, Awaitable
import asyncio

from ..message import Message


@dataclass
class InterceptorContext:
    """
    拦截器上下文
    
    Attributes:
        node_id: 当前节点 ID
        direction: 消息方向 "inbound" | "outbound"
        attributes: 拦截器间共享数据，约定的键名：
            - trace_id: 分布式追踪 ID (str)
            - auth_token: 认证令牌 (str)
            - auth_identity: 认证后的身份信息 (dict)
            - start_time: 请求开始时间 (float)
            - user_*: 用户自定义键（前缀约定）
    """
    
    node_id: str
    direction: str  # "inbound" | "outbound"
    attributes: Dict[str, Any] = field(default_factory=dict)


class Interceptor(ABC):
    """拦截器基类"""
    
    @abstractmethod
    async def intercept_inbound(
        self,
        message: Message,
        context: InterceptorContext,
        next_handler: Callable[[Message], Awaitable[Message]],
    ) -> Message:
        """
        拦截入站消息
        
        Args:
            message: 接收到的消息
            context: 拦截器上下文
            next_handler: 调用链中的下一个处理器
            
        Returns:
            处理后的消息（可修改）
            
        Raises:
            InterceptorAbort: 中止处理链
        """
        pass
    
    @abstractmethod
    async def intercept_outbound(
        self,
        message: Message,
        context: InterceptorContext,
        next_handler: Callable[[Message], Awaitable[Message]],
    ) -> Message:
        """拦截出站消息"""
        pass


class InterceptorChain:
    """
    拦截器链
    
    执行顺序遵循"洋葱模型"：
    - 添加顺序: [A, B, C]
    - 执行顺序: C.before -> B.before -> A.before -> handler -> A.after -> B.after -> C.after
    
    使用 prepend=True 可将拦截器添加到最外层（最先执行）
    """
    
    def __init__(self, interceptors: Optional[List[Interceptor]] = None):
        self._interceptors = interceptors or []
    
    def add(self, interceptor: Interceptor, prepend: bool = False) -> None:
        """
        添加拦截器
        
        Args:
            interceptor: 拦截器实例
            prepend: 是否添加到最外层（最先执行）
        """
        if prepend:
            self._interceptors.insert(0, interceptor)
        else:
            self._interceptors.append(interceptor)
    
    def remove(self, interceptor: Interceptor) -> bool:
        """移除拦截器，返回是否成功"""
        try:
            self._interceptors.remove(interceptor)
            return True
        except ValueError:
            return False
    
    def clear(self) -> None:
        """清空拦截器"""
        self._interceptors.clear()
    
    async def execute_inbound(self, message: Message, node_id: str) -> Message:
        """执行入站拦截器链"""
        context = InterceptorContext(
            node_id=node_id,
            direction="inbound",
            attributes={},
        )
        
        async def terminal(msg: Message) -> Message:
            return msg
        
        handler = terminal
        for interceptor in reversed(self._interceptors):
            handler = self._wrap_inbound(interceptor, context, handler)
        
        return await handler(message)
    
    async def execute_outbound(self, message: Message, node_id: str) -> Message:
        """执行出站拦截器链"""
        context = InterceptorContext(
            node_id=node_id,
            direction="outbound",
            attributes={},
        )
        
        async def terminal(msg: Message) -> Message:
            return msg
        
        handler = terminal
        for interceptor in reversed(self._interceptors):
            handler = self._wrap_outbound(interceptor, context, handler)
        
        return await handler(message)
    
    def _wrap_inbound(
        self,
        interceptor: Interceptor,
        context: InterceptorContext,
        next_handler: Callable[[Message], Awaitable[Message]],
    ) -> Callable[[Message], Awaitable[Message]]:
        """包装入站拦截器"""
        async def wrapped(message: Message) -> Message:
            return await interceptor.intercept_inbound(message, context, next_handler)
        return wrapped
    
    def _wrap_outbound(
        self,
        interceptor: Interceptor,
        context: InterceptorContext,
        next_handler: Callable[[Message], Awaitable[Message]],
    ) -> Callable[[Message], Awaitable[Message]]:
        """包装出站拦截器"""
        async def wrapped(message: Message) -> Message:
            return await interceptor.intercept_outbound(message, context, next_handler)
        return wrapped



"""
Trainer（服务端训练器）抽象基类
"""

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union, TYPE_CHECKING
import asyncio

from .types import TrainResult, EvalResult, ClientUpdate, RoundResult, RoundMetrics
from ..infra.logging import get_logger

if TYPE_CHECKING:
    from .model import Model
    from .aggregator import Aggregator
    from ..proxy.collection import ProxyCollection
    from ..infra.tracker import Tracker


class Trainer(ABC):
    """
    服务端训练器抽象基类

    职责：
    - 编排训练流程
    - 管理客户端（通过代理）
    - 协调聚合过程
    - 控制训练生命周期

    用户需要实现 train_round() 方法来定义单轮训练逻辑

    Callback 触发点：
    - on_train_begin: run() 开始时
    - on_train_end: run() 结束时
    - on_round_begin: train_round() 开始时
    - on_round_end: train_round() 结束时
    """

    def __init__(
        self,
        # ========== 核心依赖（必需）==========
        learners: "ProxyCollection",     # 学习器代理集合
        aggregator: "Aggregator",        # 聚合器

        # ========== 可选依赖 ==========
        dataset: Optional[Any] = None,           # 服务端数据集（用于评估，向后兼容）
        datasets: Optional[Dict[str, List[Any]]] = None,  # 数据集字典（按 split 分组）
        model: Optional["Model"] = None,         # 全局模型
        tracker: Optional["Tracker"] = None,     # 指标追踪器（所有记录通过它）
        callbacks: Optional[Any] = None,         # 回调管理器

        # ========== 配置 ==========
        config: Optional[Dict[str, Any]] = None, # 训练器配置
        node_id: Optional[str] = None,           # 节点ID（用于日志）
    ):
        """
        初始化训练器

        Args:
            learners: 学习器代理集合（ProxyCollection）
            aggregator: 聚合器
            dataset: 服务端数据集（可选，用于评估，向后兼容）
            datasets: 数据集字典，按 split 分组（可选），如 {"train": [...], "test": [...]}
            model: 全局模型（可选）
            tracker: 指标追踪器（可选，所有记录通过它）
            callbacks: 回调管理器（可选）
            config: 训练器配置
            node_id: 节点ID（用于日志）
        """
        self._learners = learners
        self._aggregator = aggregator
        self._dataset = dataset
        self._datasets = datasets or {}
        self._model = model
        self._tracker = tracker
        self._callbacks = callbacks
        self._config = config or {}
        self._node_id = node_id or "trainer"

        self.logger = get_logger(self._node_id, "training")

        # 内部状态
        self._current_round = 0
        self._is_running = False
    
    # ==================== 属性访问 ====================

    @property
    def learners(self) -> "ProxyCollection":
        """获取学习器代理集合"""
        return self._learners

    @property
    def aggregator(self) -> "Aggregator":
        """获取聚合器"""
        return self._aggregator

    @property
    def dataset(self) -> Optional[Any]:
        """获取服务端数据集"""
        return self._dataset

    @property
    def model(self) -> Optional["Model"]:
        """获取全局模型"""
        return self._model

    @property
    def tracker(self) -> Optional["Tracker"]:
        """获取追踪器"""
        return self._tracker

    @property
    def callbacks(self) -> Optional[Any]:
        """获取回调管理器"""
        return self._callbacks

    @property
    def config(self) -> Dict[str, Any]:
        """获取配置"""
        return self._config

    @property
    def current_round(self) -> int:
        """获取当前轮次"""
        return self._current_round

    @property
    def datasets(self) -> Dict[str, List[Any]]:
        """获取数据集字典"""
        return self._datasets

    @property
    def test_dataset(self) -> Optional[Any]:
        """
        获取全局测试集
        
        返回 datasets["test"] 中的第一个数据集，用于在 Trainer 端评估全局模型
        """
        test_datasets = self._datasets.get("test", [])
        return test_datasets[0] if test_datasets else None

    @property
    def has_global_test(self) -> bool:
        """是否有全局测试集"""
        return self.test_dataset is not None
    
    # ==================== 模型权重操作 ====================
    
    def get_weights(self) -> Any:
        """
        获取全局模型权重
        
        支持原生 nn.Module，返回 state_dict 格式（与 Learner 一致）
        
        Returns:
            模型权重（state_dict 格式）
        """
        if self._model is None:
            return None
        
        # 检查是否是 PyTorch 模型
        try:
            import torch.nn as nn
            if isinstance(self._model, nn.Module):
                return self._model.state_dict()
        except ImportError:
            pass
        
        # 尝试调用模型的 get_weights 方法（如果有的话）
        if hasattr(self._model, 'get_weights'):
            return self._model.get_weights()
        
        return self._model
    
    def set_weights(self, weights: Any) -> None:
        """
        设置全局模型权重
        
        支持原生 nn.Module，接收 state_dict 格式（与 Learner 一致）
        
        Args:
            weights: 模型权重（state_dict 格式）
        """
        if self._model is None:
            return
        
        # 检查是否是 PyTorch 模型
        try:
            import torch.nn as nn
            if isinstance(self._model, nn.Module):
                self._model.load_state_dict(weights, strict=False)
                return
        except ImportError:
            pass
        
        # 尝试调用模型的 set_weights 方法
        if hasattr(self._model, 'set_weights'):
            self._model.set_weights(weights)
            return
        
        self._model = weights
    
    # ==================== 核心方法（分层训练） ====================

    async def run(self) -> Dict[str, Any]:
        """
        训练会话层 - 管理整个联邦学习训练流程

        职责：
        1. 触发训练开始回调
        2. 初始化训练环境（广播初始权重）
        3. 循环执行多轮训练（调用 train_round）
        4. 触发训练结束回调

        Returns:
            训练结果字典，包含轮次结果和汇总指标

        用户可重写此方法实现自定义训练流程，或重写 train_round 实现自定义轮次逻辑
        """
        import time

        config = self._config
        max_rounds = config.get("max_rounds", 100)

        # 触发训练开始回调
        if self._callbacks:
            await self._callbacks.on_train_begin(self, {"max_rounds": max_rounds})

        # 同步训练信息（Tracker run_id 等）到所有 Learner
        await self._sync_training_info()

        # 获取在线学习器
        learners = self.get_connected_learners()
        self.logger.info(f"开始训练，客户端数量: {len(learners)}")

        # *** 新增：输出客户端状态 ***
        self._print_client_status()

        # 广播初始权重（如果有模型）
        if self._model:
            initial_weights = self.get_weights()
            await self.broadcast_to_learners("set_weights", initial_weights)

        # 训练循环
        round_results = []
        start_time = time.time()

        for round_num in range(1, max_rounds + 1):
            self._current_round = round_num

            # 计算进度信息
            progress_pct = round_num / max_rounds * 100
            elapsed = time.time() - start_time

            # 计算 ETA（从第2轮开始才有意义）
            if round_num > 1:
                avg_time_per_round = elapsed / (round_num - 1)
                eta_seconds = avg_time_per_round * (max_rounds - round_num + 1)
                eta_str = self._format_time(eta_seconds)
            else:
                eta_str = "估算中..."

            # 显示进度信息
            self.logger.info(
                f"\n{'='*70}\n"
                f"轮次 {round_num}/{max_rounds} ({progress_pct:.1f}%) | "
                f"已用时: {self._format_time(elapsed)} | 预计剩余: {eta_str}\n"
                f"{'='*70}"
            )

            # 执行单轮训练（内部会触发 on_round_begin/end）
            round_result = await self.train_round(round_num)
            round_results.append(round_result)

            # 记录轮次指标到 tracker
            # 注意：只记录评估指标（eval_*），不记录训练指标（avg_accuracy等）
            if self.tracker:
                # 过滤掉训练指标，只保留评估指标
                metrics_to_log = {
                    k: v for k, v in round_result.to_dict().items()
                    if k.startswith('eval_') or k in ['round_num', 'num_clients', 'total_samples']
                }
                if metrics_to_log:  # 只有当有评估指标时才记录
                    self.tracker.log_metrics(metrics_to_log, step=round_num)

        self.logger.info("训练完成")

        # 触发训练结束回调
        if self._callbacks:
            await self._callbacks.on_train_end(self, {"rounds_completed": max_rounds})

        # 检查配置是否需要关闭学习器
        shutdown_learners = config.get("shutdown_learners", True)  # 默认关闭学习器
        if shutdown_learners:
            self.logger.debug("发送关闭信号到所有学习器")
            try:
                # 1. 先停止心跳任务（避免在 learners 关闭后继续发送心跳）
                self._stop_heartbeat_if_needed()

                # 2. 广播关闭信号，并收集响应
                results = await self.broadcast_to_learners("_fed_shutdown", {"reason": "training_completed"})
                self.logger.debug("关闭信号已发送")

                # 3. 主动标记已响应的 learners 为断开（避免等待 gRPC channel 检测）
                self._mark_learners_disconnecting(results)

                # 4. 等待 learners 断开连接（检测连接状态，而不是固定等待时间）
                await self._wait_for_learners_disconnect(timeout=config.get("shutdown_wait_time", 10.0))
                self.logger.debug("Learners 已断开连接")
            except Exception as e:
                self.logger.warning(f"发送关闭信号失败: {e}")

        # 返回训练结果汇总
        return {
            "total_rounds": max_rounds,
            "completed_rounds": len(round_results),
            "final_round_metrics": round_results[-1].metrics.metrics if round_results else {},
        }

    @abstractmethod
    async def train_round(self, round_num: int) -> RoundResult:
        """
        轮次层 - 处理单个联邦学习轮次

        职责：
        1. 选择客户端
        2. 收集训练更新
        3. 聚合权重
        4. 广播新权重
        5. 评估（可选）
        6. 返回轮次结果

        Args:
            round_num: 轮次编号（从1开始）

        Returns:
            RoundResult: 包含更新、权重、指标

        典型实现（FedAvg）：
        ```python
        async def train_round(self, round_num: int) -> RoundResult:
            # 1. 选择客户端
            selected = self.select_learners(num_clients)

            # 2. 收集更新
            results = await self.collect_results(selected, "fit", fit_config)
            updates = [ClientUpdate.from_result(l._target_id, r)
                       for l, r in zip(selected, results) if isinstance(r, TrainResult)]

            # 3. 聚合
            new_weights = self.aggregator.aggregate(updates, self.model)
            self.model.set_weights(new_weights)

            # 4. 广播
            await self.broadcast_to_learners("set_weights", new_weights)

            # 5. 计算指标
            metrics = RoundMetrics(...)

            return RoundResult(
                round_num=round_num,
                updates=updates,
                aggregated_weights=new_weights,
                metrics=metrics
            )
        ```

        扩展点：
        - FedAvg: 标准同步聚合
        - FedProx: 添加 proximal term 到 fit_config
        - 个性化联邦学习: 只聚合部分层
        - 异步联邦学习: 不等待所有客户端
        """
        pass

    # ==================== 框架提供的方法（无需覆盖） ====================

    def get_connected_learners(self) -> List[Any]:
        """
        获取所有在线学习器（动态更新）

        使用 ProxyCollection 的动态状态，只返回当前可用的代理

        Returns:
            可用的学习器代理列表
        """
        return self._learners.get_available_proxies()

    def select_learners(
        self,
        n: int,
        strategy: str = "random",
        require_healthy: bool = False,
        **kwargs
    ) -> List[Any]:
        """
        选择学习器子集

        Args:
            n: 选择数量
            strategy: 选择策略（random等）
            require_healthy: 是否只选择健康的学习器（通过 health 状态过滤）
            **kwargs: 策略特定参数

        Returns:
            选中的学习器代理列表

        改进：
        - 支持健康检查：require_healthy=True 时只选择健康节点
        - 使用动态可用状态：只从当前在线的 learner 中选择
        """
        import random

        # 获取候选学习器（只选择可用的）
        if require_healthy:
            all_learners = self.get_healthy_learners()
            if len(all_learners) < n:
                self.logger.warning(
                    f"警告: 只有 {len(all_learners)} 个健康learner，少于请求的 {n} 个",
                    level="warning"
                )
        else:
            all_learners = self._learners.get_available_proxies()

        if n >= len(all_learners):
            return all_learners
        return random.sample(all_learners, n)

    async def broadcast_to_learners(
        self,
        method: str,
        *args,
        **kwargs
    ) -> Dict[str, Any]:
        """
        广播调用所有学习器

        Args:
            method: 方法名
            *args: 位置参数
            **kwargs: 关键字参数

        Returns:
            结果字典 {learner_id: result}
        """
        return await self._learners.broadcast(method, *args, **kwargs)

    async def broadcast_to_selected(
        self,
        learners: List[Any],
        method: str,
        *args,
        **kwargs
    ) -> List[Any]:
        """
        向指定的学习器列表广播调用

        Args:
            learners: 目标学习器列表（Proxy 对象）
            method: 方法名
            *args: 位置参数
            **kwargs: 关键字参数

        Returns:
            结果列表（与 learners 顺序对应）
        """
        if not learners:
            return []

        tasks = []
        for learner in learners:
            task = getattr(learner, method)(*args, **kwargs)
            tasks.append(asyncio.create_task(task))

        results = await asyncio.gather(*tasks, return_exceptions=True)
        return list(results)

    async def collect_results(
        self,
        learners: List[Any],
        method: str,
        *args,
        timeout: Optional[float] = None,
        min_responses: Optional[int] = None,
        **kwargs
    ) -> List[Union[TrainResult, EvalResult, Exception]]:
        """
        收集学习器结果（支持超时和部分响应）

        Args:
            learners: 学习器列表
            method: 方法名（fit / evaluate）
            *args: 位置参数
            timeout: 整体超时时间（秒），None 表示无限等待
            min_responses: 最少需要的成功响应数，达到后可以提前返回
            **kwargs: 关键字参数

        Returns:
            结果列表（成功返回 Result，失败返回 Exception或TimeoutError）

        改进：
        - 支持整体超时，避免无限等待
        - 支持最小响应数，不需要等所有learner
        - 详细的失败日志
        """
        if not learners:
            return []

        # DEBUG: 记录调用详情
        learner_ids = [getattr(l, '_target_id', 'unknown') for l in learners]
        self.logger.debug(f"collect_results: method={method}, targets={learner_ids}, args={args}, kwargs={kwargs}")

        # 创建任务列表
        tasks = []
        learner_ids = []
        for learner in learners:
            # 动态调用方法
            task = getattr(learner, method)(*args, **kwargs)
            tasks.append(asyncio.create_task(task))
            # 获取learner的ID（假设有_target_id属性）
            learner_id = getattr(learner, '_target_id', 'unknown')
            learner_ids.append(learner_id)
            self.logger.debug(f"collect_results: 创建任务, {learner_id}.{method}()")

        # 如果设置了最小响应数和超时，使用 asyncio.wait()
        if min_responses is not None or timeout is not None:
            return await self._collect_with_flexibility(
                tasks, learner_ids, min_responses, timeout
            )
        else:
            # 使用原始的 gather 方式（向后兼容）
            results = await asyncio.gather(*tasks, return_exceptions=True)
            self._log_collection_summary(learner_ids, results, method)
            return list(results)

    async def _collect_with_flexibility(
        self,
        tasks: List[asyncio.Task],
        learner_ids: List[str],
        min_responses: Optional[int],
        timeout: Optional[float]
    ) -> List[Union[TrainResult, EvalResult, Exception]]:
        """
        灵活的结果收集（支持超时和最小响应数）

        策略：
        1. 如果设置了 min_responses，达到数量后可以提前返回
        2. 如果设置了 timeout，超时后取消剩余任务
        3. 取消的任务返回 TimeoutError
        """
        num_learners = len(tasks)
        min_responses = min_responses or num_learners  # 默认等待所有响应
        results = [None] * num_learners

        try:
            # 等待任务完成
            done, pending = await asyncio.wait(
                tasks,
                timeout=timeout,
                return_when=asyncio.FIRST_EXCEPTION  # 遇到第一个异常时继续等待
            )

            # 收集已完成任务的结果
            success_count = 0
            for task in done:
                idx = tasks.index(task)
                try:
                    result = task.result()
                    results[idx] = result
                    if not isinstance(result, Exception):
                        success_count += 1
                except Exception as e:
                    results[idx] = e
                    self.logger.exception(f"Learner {learner_ids[idx]} 失败: {e}")

            # 检查是否达到最小响应数
            if success_count >= min_responses:
                self.logger.info(f"已收集到足够响应: {success_count}/{min_responses}")
            else:
                self.logger.warning(
                    f"警告: 只收集到 {success_count}/{min_responses} 个响应",
                )

            # 取消未完成的任务
            if pending:
                cancelled_ids = [learner_ids[tasks.index(t)] for t in pending]
                self.logger.warning(f"取消 {len(pending)} 个超时任务: {cancelled_ids}")
                for task in pending:
                    task.cancel()
                    idx = tasks.index(task)
                    results[idx] = TimeoutError(f"Learner {learner_ids[idx]} 超时")

        except Exception as e:
            self.logger.exception(f"收集结果时发生错误: {e}", level="error")
            # 取消所有未完成任务
            for task in tasks:
                if not task.done():
                    task.cancel()

        return results

    def _log_collection_summary(
        self,
        learner_ids: List[str],
        results: List[Any],
        method: str
    ) -> None:
        """记录收集结果摘要"""
        success_count = sum(1 for r in results if not isinstance(r, Exception))
        failed_count = len(results) - success_count

        if failed_count > 0:
            failed_ids = [
                learner_ids[i] for i, r in enumerate(results)
                if isinstance(r, Exception)
            ]
            self.logger.warning(
                f"{method} 完成: 成功 {success_count}/{len(results)}, "
                f"失败: {failed_ids}"
            )
        else:
            self.logger.info(f"{method} 完成: 全部成功 ({success_count}/{len(results)})")

    # ==================== 健康管理辅助方法 ====================

    def get_healthy_learners(self) -> List[Any]:
        """
        获取所有健康的学习器

        通过 ProxyCollection.get_healthy_proxies() 实现，
        内部查询每个 Proxy 的 health 属性（来自 Node）

        Returns:
            健康的学习器代理列表
        """
        return self._learners.get_healthy_proxies()

    async def get_learner_states(self) -> Dict[str, str]:
        """
        获取所有学习器的状态（并发查询）

        通过 ProxyCollection.get_all_states() 实现，
        每个 Proxy 内部调用远程的 state 属性

        Returns:
            {learner_id: state, ...}
        """
        return await self._learners.get_all_states()

    def get_learners_by_state(self, state: str) -> List[Any]:
        """
        按状态过滤学习器

        Args:
            state: 目标状态（idle / training / evaluating）

        Returns:
            处于该状态的学习器列表
        """
        return self._learners.get_proxies_by_state(state)

    # ==================== 辅助方法 ====================

    def _print_client_status(self):
        """
        打印客户端状态概览

        在训练开始时调用，输出所有客户端的连接和健康状态
        """
        # 获取统计信息
        stats = self._learners.get_stats()
        total_count = stats["total"]
        available_count = stats["available"]
        unavailable_count = stats["unavailable"]

        if total_count == 0:
            self.logger.warning("没有客户端！")
            return

        # 获取所有代理（包括不可用的）
        all_learners = self._learners.get_all_proxies()
        available_ids = set(stats["available_ids"])

        # 尝试获取健康状态
        try:
            healthy_learners = self.get_healthy_learners()
            healthy_ids = set(getattr(l, '_target_id', 'unknown') for l in healthy_learners)
            num_healthy = len(healthy_learners)
        except Exception:
            # 如果不支持健康检查，假设所有可用的都是健康的
            healthy_ids = available_ids
            num_healthy = available_count

        # 计算百分比
        available_pct = (available_count / total_count * 100) if total_count > 0 else 0
        health_pct = (num_healthy / total_count * 100) if total_count > 0 else 0

        # 构建输出
        separator = "=" * 70
        status_lines = [
            "",
            separator,
            "客户端状态概览",
            separator,
            f"总客户端数: {total_count}",
            f"在线客户端: {available_count} ({available_pct:.0f}%)",
            f"离线客户端: {unavailable_count}",
            f"健康客户端: {num_healthy} ({health_pct:.0f}%)",
            "",
        ]

        # 如果客户端数量不多（<=20），列出详细信息
        if total_count <= 20:
            status_lines.append("客户端列表:")

            for learner in all_learners:
                learner_id = getattr(learner, '_target_id', 'unknown')
                is_available = learner_id in available_ids
                is_healthy = learner_id in healthy_ids

                # 状态图标
                if is_available and is_healthy:
                    status_icon = "✓"
                    status_str = "ONLINE/HEALTHY"
                elif is_available:
                    status_icon = "⚠"
                    status_str = "ONLINE/UNHEALTHY"
                else:
                    status_icon = "✗"
                    status_str = "OFFLINE"

                status_lines.append(f"  {status_icon} {learner_id} [{status_str}]")
        else:
            # 客户端太多，只显示统计信息
            status_lines.append(f"（共 {total_count} 个客户端，详细列表已省略）")

        status_lines.append(separator)
        status_lines.append("")

        # 输出状态信息
        status_msg = "\n".join(status_lines)
        self.logger.info(status_msg)

        # 警告信息
        if unavailable_count > 0:
            self.logger.warning(
                f"检测到 {unavailable_count} 个离线客户端，它们不会参与训练"
            )

        if num_healthy < available_count:
            unhealthy_count = available_count - num_healthy
            self.logger.warning(
                f"检测到 {unhealthy_count} 个不健康的在线客户端，可能影响训练质量"
            )

    def _format_time(self, seconds: float) -> str:
        """
        格式化时间显示

        Args:
            seconds: 秒数

        Returns:
            格式化的时间字符串（如 "1h 23m 45s" 或 "5m 30s"）
        """
        if seconds < 60:
            return f"{int(seconds)}秒"
        elif seconds < 3600:
            minutes = int(seconds / 60)
            secs = int(seconds % 60)
            return f"{minutes}分{secs}秒"
        else:
            hours = int(seconds / 3600)
            minutes = int((seconds % 3600) / 60)
            return f"{hours}小时{minutes}分"

    def _mark_learners_disconnecting(self, results: Dict[str, Any]):
        """
        主动标记已响应 shutdown 的 learners 为断开状态

        Args:
            results: broadcast_to_learners 返回的结果字典 {learner_id: result}

        此方法直接修改 node_comm 的连接表，避免等待 gRPC channel 检测断开
        """
        # 从 ProxyCollection 获取 node
        node = self._learners._node if hasattr(self._learners, '_node') else None
        if not node:
            self.logger.debug("无法获取 node 引用，跳过标记断开")
            return

        # 获取 node_comm 实例
        comm = node._comm if hasattr(node, '_comm') else None
        if not comm:
            self.logger.debug("无法获取 comm 引用，跳过标记断开")
            return

        # 获取连接表
        connection_table = comm._connection_table if hasattr(comm, '_connection_table') else None
        if connection_table is None:
            self.logger.debug("无法获取连接表，跳过标记断开")
            return

        # 导入 ConnectionStatus
        from ..comm.message import ConnectionStatus

        # 统计成功响应的 learners
        success_count = 0
        for learner_id, result in results.items():
            # 检查是否成功响应（不是异常）
            if not isinstance(result, Exception):
                # 检查连接表中是否有此 learner
                if learner_id in connection_table:
                    # 标记为断开
                    connection_table[learner_id].status = ConnectionStatus.DISCONNECTED
                    success_count += 1
                    self.logger.debug(f"已标记 {learner_id} 为断开状态")

        if success_count > 0:
            self.logger.debug(f"已主动标记 {success_count} 个 learners 为断开状态")

    def _stop_heartbeat_if_needed(self):
        """
        停止心跳任务（如果使用 gRPC 传输）

        在关闭 learners 之前调用，避免在 learners 关闭后继续发送心跳导致 gRPC 错误
        """
        # 从 ProxyCollection 获取 node
        node = self._learners._node if hasattr(self._learners, '_node') else None
        if not node:
            self.logger.debug("无法获取 node 引用，跳过停止心跳")
            return

        # 获取 transport 层
        comm = node._comm if hasattr(node, '_comm') else None
        if not comm:
            self.logger.debug("无法获取 comm 引用，跳过停止心跳")
            return

        transport = comm._transport if hasattr(comm, '_transport') else None
        if not transport:
            self.logger.debug("无法获取 transport 引用，跳过停止心跳")
            return

        # 检查是否是 gRPC 传输且有控制线程
        if hasattr(transport, '_control_running') and transport._control_running:
            self.logger.debug("停止 gRPC 心跳任务")
            transport._stop_control_thread()
            self.logger.debug("gRPC 心跳任务已停止")

    async def _wait_for_learners_disconnect(self, timeout: float = 10.0):
        """
        等待所有 learners 断开连接

        Args:
            timeout: 超时时间（秒）

        此方法会轮询检查连接状态，直到所有 learners 断开或超时
        """
        import time
        start_time = time.time()
        check_interval = 0.1  # 100ms 检查一次

        # 从 ProxyCollection 获取 node
        node = self._learners._node if hasattr(self._learners, '_node') else None
        if not node:
            self.logger.warning("无法获取 node 引用，跳过等待 learners 断开")
            return

        learner_ids = [proxy._target_id for proxy in self._learners]

        while time.time() - start_time < timeout:
            # 获取当前连接的节点
            connected_nodes = node.get_connected_nodes()

            # 检查是否还有 learners 连接
            still_connected = [lid for lid in learner_ids if lid in connected_nodes]

            if not still_connected:
                # 所有 learners 都断开了
                return

            # 等待一小段时间再检查
            await asyncio.sleep(check_interval)

        # 超时，记录调试信息（这在 gRPC 中是正常的，因为 channel 断开检测有延迟）
        connected_nodes = node.get_connected_nodes()
        still_connected = [lid for lid in learner_ids if lid in connected_nodes]
        if still_connected:
            self.logger.debug(
                f"等待 learners 断开超时（{timeout}秒），"
                f"仍有 {len(still_connected)} 个 learners 连接: {still_connected}"
            )

    async def _sync_training_info(self):
        """
        同步训练信息到所有 Learner（Trainer端）

        遍历所有 SyncCallback，收集同步信息并广播
        """
        if not self._callbacks:
            return

        from ..callback.sync_callback import SyncCallback

        sync_info = {}

        # 收集所有 SyncCallback 的同步信息
        for callback in self._callbacks._callbacks:
            if isinstance(callback, SyncCallback):
                try:
                    info = await callback.collect_sync_info()
                    sync_info.update(info)
                except Exception as e:
                    self.logger.error(f"Failed to collect sync info from {callback}: {e}")

        # 如果有同步信息，广播给所有 Learner
        if sync_info:
            self.logger.info(f"Broadcasting training sync info: {list(sync_info.keys())}")
            try:
                await self.broadcast_to_learners("sync_training_info", sync_info)
                self.logger.info("Training sync info broadcasted successfully")
            except Exception as e:
                self.logger.error(f"Failed to broadcast sync info: {e}")

    def increment_round(self) -> int:
        """增加轮次计数"""
        self._current_round += 1
        return self._current_round


"""
Transport 工厂
"""

from typing import TYPE_CHECKING

from .base import Transport
from .memory import MemoryTransport
from .grpc_transport import GrpcTransport

# 使用 config/schema.py 中的配置类
from ...config import TransportConfig

if TYPE_CHECKING:
    from ..node import Node


def create_transport(config: TransportConfig, node: "Node") -> Transport:
    """
    根据配置创建 Transport 实例
    
    Args:
        config: 传输层配置 (oiafed.config.TransportConfig)
        node: 所属节点
        
    Returns:
        Transport 实例
    """
    if config.mode == "memory":
        return MemoryTransport(node, config.memory)
    elif config.mode == "grpc":
        return GrpcTransport(node, config.grpc)
    else:
        raise ValueError(f"Unknown transport mode: {config.mode}")



"""
FedYogi 聚合器

从 methods/aggregators/fedyogi.py 迁移到 src/

实现 FedYogi 聚合算法,基于Yogi优化器的联邦学习版本。
提供比FedAdam更稳定的自适应聚合。

论文: Adaptive Federated Optimization
作者: Sashank J. Reddi et al.
发表: ICLR 2021
"""

import torch
from typing import List, Dict, Any
from loguru import logger

from ...core.aggregator import Aggregator
from ...core.types import ClientUpdate
from ...registry import aggregator


@aggregator(
    name='fedyogi',
    description='FedYogi自适应联邦聚合器',
    version='1.0'
)
class FedYogiAggregator(Aggregator):
    """FedYogi 聚合器实现"""

    def __init__(self, server_lr: float = 1e-2, beta1: float = 0.9,
                 beta2: float = 0.99, eps: float = 1e-3, **kwargs):
        """初始化FedYogi聚合器"""
        self.server_lr = server_lr
        self.beta1 = beta1
        self.beta2 = beta2
        self.eps = eps

        device = kwargs.get("device", "auto")
        if device == "auto":
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self.device = device

        # Yogi状态
        self.m_state = {}
        self.v_state = {}
        self.round_count = 0

        # 全局模型权重(需要用于计算梯度)
        self.global_weights = None

        logger.info(f"✅ FedYogi聚合器初始化完成 - LR: {self.server_lr}")

    def aggregate(self, updates: List[ClientUpdate], global_model=None) -> Dict[str, torch.Tensor]:
        """
        执行FedYogi聚合

        Args:
            updates: 客户端更新列表 (List[ClientUpdate])
            global_model: 全局模型 (可选)

        Returns:
            聚合后的模型权重字典
        """
        if not updates:
            raise ValueError("没有客户端更新可聚合")

        self.round_count += 1

        # 初始化全局权重(第一轮)
        if self.global_weights is None:
            first_weights = updates[0].weights
            self.global_weights = {}
            for k, v in first_weights.items():
                v_tensor = v.clone().to(self.device)
                # 转换为float以便后续计算
                if v_tensor.dtype in [torch.long, torch.int, torch.int32, torch.int64]:
                    v_tensor = v_tensor.float()
                self.global_weights[k] = v_tensor

        # 计算伪梯度: Δ = weighted_avg(client_weights) - global_weights
        aggregated_delta = self._compute_pseudo_gradient(updates)

        # 初始化Yogi状态
        if not self.m_state:
            self._initialize_yogi_states(aggregated_delta)

        # Yogi更新全局权重
        self.global_weights = self._yogi_update(aggregated_delta)

        # 转换回原始类型
        result = {}
        first_weights = updates[0].weights
        for param_name, param_value in self.global_weights.items():
            if first_weights[param_name].dtype in [torch.long, torch.int, torch.int32, torch.int64]:
                result[param_name] = param_value.long()
            else:
                result[param_name] = param_value

        return result

    def _compute_pseudo_gradient(self, updates: List[ClientUpdate]) -> Dict[str, torch.Tensor]:
        """
        计算伪梯度: Δ = Σ(w_i * client_weights_i) - global_weights
        这代表客户端权重加权平均与当前全局权重的差值
        """
        total_samples = sum(update.num_samples for update in updates)
        weighted_avg = {}

        first_weights = updates[0].weights
        for param_name in first_weights:
            # 使用float32避免Long类型转换错误
            weighted_avg[param_name] = torch.zeros_like(
                first_weights[param_name], dtype=torch.float32, device=self.device
            )

            # 计算加权平均
            for update in updates:
                weight = update.num_samples / total_samples
                param_value = update.weights[param_name].to(self.device)
                # 转换整数类型为float
                if param_value.dtype in [torch.long, torch.int, torch.int32, torch.int64]:
                    param_value = param_value.float()
                weighted_avg[param_name] += weight * param_value

            # 如果原始参数是整数类型，转换回去
            if first_weights[param_name].dtype in [torch.long, torch.int, torch.int32, torch.int64]:
                weighted_avg[param_name] = weighted_avg[param_name].long()

        # 计算伪梯度: Δ = weighted_avg - global_weights
        pseudo_gradient = {}
        for param_name in weighted_avg:
            pseudo_gradient[param_name] = weighted_avg[param_name] - self.global_weights[param_name]

        return pseudo_gradient

    def _initialize_yogi_states(self, gradient: Dict[str, torch.Tensor]):
        """初始化Yogi状态"""
        for param_name, grad in gradient.items():
            self.m_state[param_name] = torch.zeros_like(grad, device=self.device)
            self.v_state[param_name] = torch.zeros_like(grad, device=self.device)

    def _yogi_update(self, delta: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        执行Yogi更新
        global_weights = global_weights + server_lr * m / (sqrt(|v|) + eps)
        """
        updated_weights = {}

        for param_name, grad in delta.items():
            # 更新一阶矩
            self.m_state[param_name] = self.beta1 * self.m_state[param_name] + (1 - self.beta1) * grad

            # Yogi的二阶矩更新(与Adam不同)
            grad_squared = grad ** 2
            v_diff = grad_squared - self.v_state[param_name]
            self.v_state[param_name] = self.v_state[param_name] + (1 - self.beta2) * torch.sign(v_diff) * v_diff

            # 偏置修正
            m_corrected = self.m_state[param_name] / (1 - self.beta1 ** self.round_count)
            v_corrected = self.v_state[param_name] / (1 - self.beta2 ** self.round_count)

            # Yogi更新: w = w + lr * m / (sqrt(|v|) + eps)
            updated_weights[param_name] = (
                self.global_weights[param_name] +
                self.server_lr * m_corrected / (torch.sqrt(torch.abs(v_corrected)) + self.eps)
            )

        return updated_weights

    def get_stats(self) -> Dict[str, Any]:
        """获取聚合器统计信息"""
        return {
            "algorithm": "FedYogi",
            "server_lr": self.server_lr,
            "rounds": self.round_count
        }
